#ifndef AVS_H
#define AVS_H

#include "lectura.h"

/**
 * @file avs.h
 * @brief Funciones para escribir ficheros AVS
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


/**
 * @brief Función que graba en un fichero el campo de viento resultante
 *        con un formato adecuado para ser visualizado con el AVS
 * @param[in] fich Nombre del fichero a grabar
 * @param[in] malla Malla donde hay la información del viento
 */
void 
GrabarVientoAVS(const char *fich, 
        const struct T_Malla *malla);

/**
 * @brief Función que graba en un fichero el módulo del campo de viento resultante
 *        con un formato adecuado para ser visualizado con el AVS
 * @param[in] fich Nombre del fichero a grabar
 * @param[in] malla Malla donde hay la información del viento
 */

void GrabarModuloVientoAVS(const char *fich,
                           const struct T_Malla *malla);

/**
 * @brief Función que graba en un fichero el campo de viento interpolado
 *        con un formato adecuado para ser visualizado con el AVS
 * @param[in] fich Nombre del fichero a grabar
 * @param[in] malla Malla donde hay la información del viento
 */
void GrabarVientoInter(const char *fich, const struct T_Malla *malla);


/**
 * @brief Función que busca el tetraedro de la malla al que pertenece un punto
 * @param[in] malla Malla en la que buscar el tetraedro
 * @param[in] X,Y,Z Coordenada X Y Z del punto
 * @return Devuelve el número del tetraedro al que pertenece el punto
 *         o -1 en caso que el punto no pertenzca a ningún tetraedro
 */
int
BuscaPuntoEnMalla(const struct T_Malla  *malla, const double *X, const double *Y, const double *Z);


/**
 * @brief Función que calcula la distancia entre dos nodos cualquiera de la malla
 * @param[in] malla Estructura de la malla
 * @param[in] nodo1 Número de referencia del primer nodo en la malla
 * @param[in] nodo2 Número de referencia del segundo nodo en la malla
 * @return Distancia entre los dos nodos de la malla
 */
double
DistanciaEntreNodos(const struct T_Malla *malla, const size_t nodo1, const size_t nodo2);

/**
 * @brief Función que calcula el volumen de un tetraedro
 * @param[in] x1 Coordenada X del nodo 1
 * @param[in] y1 Coordenada Y del nodo 1
 * @param[in] z1 Coordenada Z del nodo 1
 * @param[in] x2 Coordenada X del nodo 2
 * @param[in] y2 Coordenada Y del nodo 2
 * @param[in] z2 Coordenada Z del nodo 2
 * @param[in] x3 Coordenada X del nodo 3
 * @param[in] y3 Coordenada Y del nodo 3
 * @param[in] z3 Coordenada Z del nodo 3
 * @param[in] x4 Coordenada X del nodo 4
 * @param[in] y4 Coordenada Y del nodo 4
 * @param[in] z4 Coordenada Z del nodo 4
 * @return Volumen del tetraedro
 */
double VolTetra(const double* x1, const double* y1, const double* z1,
                const double* x2, const double* y2, const double* z2,
                const double* x3, const double* y3, const double* z3,
                const double* x4, const double* y4, const double* z4);


#endif // AVS_H
