#include "calculo.h"

#include <math.h>
#include <stdlib.h>

/**
 * @file calculo.c
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @brief Derivadas de las funciones de forma
 */

const double dff[4][3] = {
    { -1.0, -1.0, -1.0 } ,
    {  1.0,  0.0,  0.0 } ,
    {  0.0,  1.0,  0.0 } ,
    {  0.0,  0.0,  1.0 }
};


double djac(double A[3][3])
{
    int ncont=0;
    int ir, ifil, icol, i, j, k;
    double amax, aux;

    // Determinacion del pivote maximo
    for (ir=0; ir < 2; ir++) {
        amax = fabs(A[ir][ir]);
        ifil = icol = ir;
        for (i=ir; i < 3; i++) {
            for (j=ir; j < 3; j++) {
                if (fabs(A[i][j]) > amax) {
                    ifil = i;
                    icol = j;
                    amax = fabs(A[i][j]);
                }
            }
        }
        // Intercambio de la fila ir por la fila ifil
        if (ifil != ir) {
            ncont++;
            for (k=ir; k < 3; k++) {
                aux = A[ir][k];
                A[ir][k] =  A[ifil][k];
                A[ifil][k] = aux;
            }
        }
        // Intercambio de columnas
        if (icol != ir) {
            ncont++;
            for (k=0; k < 3; k++) {
                aux = A[k][ir];
                A[k][ir] = A[k][icol];
                A[k][icol] = aux;
            }
        }

        for (i=ir+1; i < 3; i++) {
            for (j=ir+1; j < 3; j++) {
                A[i][j] = A[i][j] - A[i][ir] * A[ir][j] / A[ir][ir];
            }
        }
    }

    aux = A[0][0] * A[1][1] * A[2][2];
    if (ncont == 1 || ncont == 3) {
        aux = -aux;
    }
    return(aux);
} // djac

void
Elemento(const size_t inel, const struct T_Malla *malla, struct T_MElemental *A, double *b,
         const double *alpha)
{
    double cj[3][3], cji[3][3], xpi[3];
    double dj;
    double a1, a2, a3, b1, b2, b3;
    int i, j, k;

    double ff[4] = { 0.25, 0.25, 0.25, 0.25 }; // Funciones de forma
    double peso = 1.0 /6.0; // Peso de las funciones de forma

    for(i=0;i<3;i++){
        for(j=0;j<3;j++){
            cj[i][j] = 0.0;
            for(k=0; k<4; k++){
                cj[i][j]= cj[i][j] + malla->nodo[malla->tetra[inel][k]][i] * dff[k][j];
                cji[i][j] = cj[i][j];
            }
        }
    }




    for(i=0;i<3;i++){
        xpi[i] = 0.0;
        for(k=0;k<4;k++) {
            xpi[i]+= malla->nodo[malla->tetra[inel][k]][i] * ff[k];
        }
    }

    dj = djac(cji);
    if(dj < 0.0){
        fprintf(stderr,
                "Elemento:jacobiano de distinto signo (elto. %zu)\n",
                inel);
        exit(EXIT_FAILURE);
    }

    cji[0][0] = cj[1][1] * cj[2][2] - cj[2][1] * cj[1][2];
    cji[0][1] = cj[2][1] * cj[0][2] - cj[0][1] * cj[2][2];
    cji[0][2] = cj[0][1] * cj[1][2] - cj[1][1] * cj[0][2];

    cji[1][0] = cj[2][0] * cj[1][2] - cj[1][0] * cj[2][2];
    cji[1][1] = cj[0][0] * cj[2][2] - cj[2][0] * cj[0][2];
    cji[1][2] = cj[1][0] * cj[0][2] - cj[0][0] * cj[1][2];

    cji[2][0] = cj[1][0] * cj[2][1] - cj[2][0] * cj[1][1];
    cji[2][1] = cj[2][0] * cj[0][1] - cj[0][0] * cj[2][1];
    cji[2][2] = cj[0][0] * cj[1][1] - cj[1][0] * cj[0][1];

    for(i=0;i<4; i++){
        b[i] = 0.0;
        for(j=0; j<4; j++) {
            A->valor[i][j] = 0.0;
        }
    }



    for(i=0;i<4; i++){
        a1 = a2 = a3 = 0.0;
        A->indice[i] = malla->tetra[inel][i];
        for(k=0; k<3; k++){
            a1 = a1 + dff[i][k] * cji[k][0];
            a2 = a2 + dff[i][k] * cji[k][1];
            a3 = a3 + dff[i][k] * cji[k][2];
        }
        for(j=0; j<4; j++){
            b1 = b2 = b3 = 0.0;
            for(k=0; k<3; k++){
                b1+= dff[j][k] * cji[k][0];
                b2+= dff[j][k] * cji[k][1];
                b3+= dff[j][k] * cji[k][2];
            }
            A->valor[i][j]+= (a1*b1 + a2*b2 +(*alpha)*(*alpha)*a3*b3);
            A->valor[i][j]/= dj;
            A->valor[i][j]*= peso;
        }

        b[i] -= (a1 * malla->vini[inel][0] + a2 * malla->vini[inel][1] +
                a3 * malla->vini[inel][2]) * peso;
    }
}


void CalculaViento(struct T_Malla *malla, const double *x, const double *alpha)
{
    double cj[3][3], cji[3][3];

    //    malla->ind_err_max = -1.0;
    //    alpha = datmos->alpha * datmos->alpha;
    for(size_t inel=0; inel < malla->Num_tetra; inel++){
        malla->wind[inel][0] = malla->wind[inel][1] = malla->wind[inel][2] = 0.0;
        for(size_t i=0;i<3;i++){
            for(size_t j=0;j<3;j++){
                cj[i][j] = 0.0;
                for(size_t k=0; k<4; k++){
                    cj[i][j]+= malla->nodo[malla->tetra[inel][k]][i] * dff[k][j];
                    cji[i][j] = cj[i][j];
                }
            }
        }


        double dj = djac(cji);

        cji[0][0] = cj[1][1] * cj[2][2] - cj[2][1] * cj[1][2];
        cji[0][1] = cj[2][1] * cj[0][2] - cj[0][1] * cj[2][2];
        cji[0][2] = cj[0][1] * cj[1][2] - cj[1][1] * cj[0][2];

        cji[1][0] = cj[2][0] * cj[1][2] - cj[1][0] * cj[2][2];
        cji[1][1] = cj[0][0] * cj[2][2] - cj[2][0] * cj[0][2];
        cji[1][2] = cj[1][0] * cj[0][2] - cj[0][0] * cj[1][2];

        cji[2][0] = cj[1][0] * cj[2][1] - cj[2][0] * cj[1][1];
        cji[2][1] = cj[2][0] * cj[0][1] - cj[0][0] * cj[2][1];
        cji[2][2] = cj[0][0] * cj[1][1] - cj[1][0] * cj[0][1];

        for(size_t i=0;i<4; i++){
            double a1 = 0;
            double a2 = 0;
            double a3 = 0;

            for(size_t k=0; k<3; k++){
                a1+= dff[i][k] * cji[k][0];
                a2+= dff[i][k] * cji[k][1];
                a3+= dff[i][k] * cji[k][2];
            }

            malla->wind[inel][0]+= a1*x[malla->tetra[inel][i]];
            malla->wind[inel][1]+= a2*x[malla->tetra[inel][i]];
            malla->wind[inel][2]+= a3*x[malla->tetra[inel][i]];
        }

        for(size_t i=0;i<3; i++) {
            malla->wind[inel][i] /= dj;
        }

        malla->wind[inel][0]+= malla->vini[inel][0];
        malla->wind[inel][1]+= malla->vini[inel][1];
        malla->wind[inel][2] = (((*alpha)*(*alpha)) * malla->wind[inel][2] +
                malla->vini[inel][2]);
    }
}// CalculaViento
