#ifndef CALCULO_H
#define CALCULO_H

#include "ensamblaje.h"

/**
 * @file calculo.h
 * @brief Módulo de cálculo para el problema de masa consistente
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @brief Función que Dado el elemento inel-esimo de la malla,
 *        calcula la matriz y el vector elementales asociados.
 * @param[in] inel Número del elemento de la malla
 * @param[in] malla Malla
 * @param[out] A Matriz elemental resultante (4x4)
 * @param[out] b Vector elemental de terminos independientes resultante
 * @param[in] alpha Parámetro alpha del modelo
 */
void
Elemento(const size_t inel, const struct T_Malla *malla, struct T_MElemental *A, double *b,
     const double *alpha);


/**
 * @brief Función que Calcula el viento a partir de la solucion del sistema de ecuaciones,
 *        las funciones de forma y sus derivadas.
 * @param[out] malla Malla donde calcular el viento, se modifica el miembro wind
 * @param[in] x Solción del sistema de ecuaciones
 * @param[in] alpha Parámetro alpha del modelo
 */
void
CalculaViento(struct T_Malla *malla, const double *x, const double *alpha);


/**
 * @brief Función que calcula el determinante de una matriz de 3x3 con pivotacion
 *        para evitar problemas con tetraedros estirados
 * @param[inout] A Matriz de la que calcular el determinante, se modifica con el método del pivote máximo
 * @return Determinante de la matriz A
 */
double djac(double A[3][3]);


#endif //CALCULO_H
