#include "capas.h"

#include "avs.h"
#include "localizador.h"
#include "viento.h"

#include <math.h>
#include <stdlib.h>

/**
 * @file capas.c
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

void calculaCapas(const struct T_Malla *malla, const struct T_DatosAtmos *Datmos,
        const struct T_Topografia *topo, const struct T_NEstacion *todas_est, const char *file_format)
{

    size_t h_tam = 9;

    double *h = (double *)malloc( h_tam * sizeof(double));
    h[0] = 10;
    h[1] = 20;
    h[2] = 50;
    h[3] = 100;
    h[4] = 150;
    h[5] = 200;
    h[6] = 300;
    h[7] = 400;
    h[8] = 500;

	struct sCapaTerreno terreno_capa;

	terreno_capa.nx = 30;
	terreno_capa.ny = 30;

	double dx = (malla->terreno.x_max - malla->terreno.x_min)/terreno_capa.nx;
	double dy = (malla->terreno.y_max - malla->terreno.y_min)/terreno_capa.ny;


	terreno_capa.nodos = (double **)malloc( (terreno_capa.nx*terreno_capa.ny) * sizeof(double *));
	for (size_t i=0; i < terreno_capa.nx; i++) {
		for(size_t j=0; j<terreno_capa.ny; j++) {
			terreno_capa.nodos[j+terreno_capa.nx*i] = (double *)malloc(3*sizeof(double));
			terreno_capa.nodos[j+terreno_capa.nx*i][0] = malla->terreno.x_min + dx/2.0 + i*dx;
			terreno_capa.nodos[j+terreno_capa.nx*i][1] = malla->terreno.y_min + dy/2.0 + j*dy;
			terreno_capa.nodos[j+terreno_capa.nx*i][2] = 0.0;
		}
	}

	
	for (size_t i=0; i < terreno_capa.nx; i++) {
		for(size_t j=0; j<terreno_capa.ny; j++) {

			double *punto = (double *)malloc(3*sizeof(double));
			punto[0]= terreno_capa.nodos[j+terreno_capa.nx*i][0];
			punto[1]= terreno_capa.nodos[j+terreno_capa.nx*i][1];
			punto[2]= terreno_capa.nodos[j+terreno_capa.nx*i][2];

            terreno_capa.nodos[j+terreno_capa.nx*i][2]=BuscarAltura(punto, malla);

			free(punto);



		}
	}



	for (size_t h_idx = 0; h_idx < h_tam; h_idx++) {

		struct sVientoCapa capa_viento;
		struct sVientoCapa capa_viento_ini;

        capa_viento.wind = (struct T_Vector3D *)malloc(
				(terreno_capa.nx*terreno_capa.ny) *
                sizeof(struct T_Vector3D));

        capa_viento_ini.wind = (struct T_Vector3D *)malloc(
				(terreno_capa.nx*terreno_capa.ny) *
                sizeof(struct T_Vector3D));

		for (size_t i=0; i<terreno_capa.nx; i++) {
			for (size_t j=0; j<terreno_capa.ny; j++) {
				size_t idx_terr = j+terreno_capa.nx*i;
                const double height = terreno_capa.nodos[idx_terr][2]+h[h_idx];

                size_t tetra = getTetrahedra(terreno_capa.nodos[idx_terr][0],
                                          terreno_capa.nodos[idx_terr][1],
                                          height,
                                          (const double**)malla->nodo, (const size_t **)malla->tetra, malla->Num_tetra, &malla->situation);

                if ( tetra == malla->Num_tetra+1) {
					fprintf(stdout, "No se ha encontrado tetraedro\n");
                    fprintf(stdout, "h: %f\ni: %zu\npunto: %f %f %f\nsuma altura: %f\n",
							h[h_idx], i,
							terreno_capa.nodos[idx_terr][0],
							terreno_capa.nodos[idx_terr][1],
							terreno_capa.nodos[idx_terr][2],
							terreno_capa.nodos[idx_terr][2]+h[h_idx]);
					fprintf(stderr, "No se ha encontrado tetraedro\n");
                    exit(EXIT_FAILURE);
				}

				capa_viento.wind[idx_terr] = VientoEnPunto(malla,
						topo, Datmos, todas_est,
						tetra,
						terreno_capa.nodos[idx_terr][0],
						terreno_capa.nodos[idx_terr][1],
                        terreno_capa.nodos[idx_terr][2]+h[h_idx],
                        terreno_capa.nodos[idx_terr][2],
                        file_format,&(capa_viento_ini.wind[idx_terr]));
			}
		}

		char *fichero;

		int len = snprintf(NULL, 0, "%s%g%s", "viento_capa_vect_", h[h_idx],
				".inp");
		fichero = (char *)malloc((len + 1)* sizeof(char));
		snprintf(fichero, len+1, "%s%g%s", "viento_capa_vect_", h[h_idx], ".inp");

		FILE *fp=fopen(fichero, "w");
		if (!fp) {
			fprintf(stderr, "Problema al intentar abrir el archivo %s", fichero);
            exit(EXIT_FAILURE);
		}

        fprintf(fp, "%zu %zu 3 0 0\n", terreno_capa.nx * terreno_capa.ny, terreno_capa.nx * terreno_capa.ny);

		for (size_t i=0; i<terreno_capa.nx; i++) {
			for (size_t j=0; j<terreno_capa.ny; j++) {
				size_t idx_terr = j+terreno_capa.nx*i;
                fprintf(fp, " %zu %f %f %f\n", idx_terr+1,
						terreno_capa.nodos[idx_terr][0],
						terreno_capa.nodos[idx_terr][1],
						terreno_capa.nodos[idx_terr][2]+h[h_idx]);
			}
		}

		for (size_t i=0; i < terreno_capa.nx * terreno_capa.ny; i++) {
            fprintf (fp, "%zu %d %s %zu\n", i+1, 0, " pt ", i+1);
		}

		fprintf(fp, "1 3\nviento, ms-1\n");

		for (size_t i=0; i<terreno_capa.nx; i++) {
			for (size_t j=0; j<terreno_capa.ny; j++) {
				size_t idx_terr = j+terreno_capa.nx*i;
                fprintf (fp, "%zu %f %f %f\n", idx_terr+1, capa_viento.wind[idx_terr].x,
						capa_viento.wind[idx_terr].y,
						capa_viento.wind[idx_terr].z);
			}
		}

		int iostat = fclose(fp);
		if (iostat != 0) {
			fprintf(stderr, "Problema al cerrar el archivo %s", fichero);
            exit(EXIT_FAILURE);
		}

		free(fichero);

		len = snprintf(NULL, 0, "%s%g%s", "viento_capa_ini_vect_", h[h_idx],
				".inp");
		fichero = (char *)malloc((len + 1)* sizeof(char));
		snprintf(fichero, len+1, "%s%g%s", "viento_capa_ini_vect_", h[h_idx], ".inp");

		fp=fopen(fichero, "w");
		if (!fp) {
			fprintf(stderr, "Problema al intentar abrir el archivo %s", fichero);
            exit(EXIT_FAILURE);
		}

        fprintf(fp, "%zu %zu 3 0 0\n", terreno_capa.nx * terreno_capa.ny, terreno_capa.nx * terreno_capa.ny);

		for (size_t i=0; i<terreno_capa.nx; i++) {
			for (size_t j=0; j<terreno_capa.ny; j++) {
				size_t idx_terr = j+terreno_capa.nx*i;
                fprintf(fp, " %zu %f %f %f\n", idx_terr+1,
						terreno_capa.nodos[idx_terr][0],
						terreno_capa.nodos[idx_terr][1],
						terreno_capa.nodos[idx_terr][2]+h[h_idx]);
			}
		}

		for (size_t i=0; i < terreno_capa.nx * terreno_capa.ny; i++) {
            fprintf (fp, "%zu %d %s %zu\n", i+1, 0, " pt ", i+1);
		}

		fprintf(fp, "1 3\nviento, ms-1\n");

		for (size_t i=0; i<terreno_capa.nx; i++) {
			for (size_t j=0; j<terreno_capa.ny; j++) {
				size_t idx_terr = j+terreno_capa.nx*i;
                fprintf (fp, "%zu %f %f %f\n", idx_terr+1, capa_viento_ini.wind[idx_terr].x,
						capa_viento_ini.wind[idx_terr].y,
						capa_viento_ini.wind[idx_terr].z);
			}
		}

		iostat = fclose(fp);
		if (iostat != 0) {
			fprintf(stderr, "Problema al cerrar el archivo %s", fichero);
            exit(EXIT_FAILURE);
		}

		free(fichero);

		len = snprintf(NULL, 0, "%s%g%s", "viento_capa_esc_", h[h_idx],
				".inp");
		fichero = (char *)malloc((len + 1)* sizeof(char));
		snprintf(fichero, len+1, "%s%g%s", "viento_capa_esc_", h[h_idx], ".inp");

		fp=fopen(fichero, "w");
		if (!fp) {
			fprintf(stderr, "Problema al intentar abrir el archivo %s", fichero);
            exit(EXIT_FAILURE);
		}

        fprintf(fp, "%zu %zu 1 0 0\n", terreno_capa.nx * terreno_capa.ny, ((terreno_capa.nx-1) * (terreno_capa.ny-1) * 2));

		for (size_t i=0; i<terreno_capa.nx; i++) {
			for (size_t j=0; j<terreno_capa.ny; j++) {
				size_t idx_terr = j+terreno_capa.nx*i;
                fprintf(fp, " %zu %f %f %f\n", idx_terr+1,
						terreno_capa.nodos[idx_terr][0],
						terreno_capa.nodos[idx_terr][1],
						terreno_capa.nodos[idx_terr][2]+h[h_idx]);
			}
		}

		{size_t k = 1;
			for (size_t i=0; i < (terreno_capa.nx-1) ; i++) {
				for (size_t j=0; j < (terreno_capa.ny-1); j++) {
                    fprintf (fp, "%zu %d %s %zu %zu %zu\n",
							k, 0, " tri ",
							(j*terreno_capa.nx)+i+1,
							(j*terreno_capa.nx)+i+2,((j+1)*terreno_capa.nx)+i+2);
					k++;
                    fprintf (fp, "%zu %d %s %zu %zu %zu\n",
							k, 0, " tri ",
							(j*terreno_capa.nx)+i+1,
							((j+1)*terreno_capa.nx)+i+2,((j+1)*terreno_capa.nx)+i+1);
					k++;
				}
			}
		}

		fprintf(fp, "1 1\nviento, ms-1\n");

		for (size_t i=0; i<terreno_capa.nx; i++) {
			for (size_t j=0; j<terreno_capa.ny; j++) {
				size_t idx_terr = j+terreno_capa.nx*i;
            fprintf (fp, "%zu %f \n", idx_terr+1, sqrt((capa_viento_ini.wind[idx_terr].x*capa_viento_ini.wind[idx_terr].x)+
					(capa_viento_ini.wind[idx_terr].y*capa_viento_ini.wind[idx_terr].y)+
					(capa_viento_ini.wind[idx_terr].z*capa_viento_ini.wind[idx_terr].z)));
			}
		}

		iostat = fclose(fp);
		if (iostat != 0) {
			fprintf(stderr, "Problema al cerrar el archivo %s", fichero);
            exit(EXIT_FAILURE);
		}

		free(fichero);

		len = snprintf(NULL, 0, "%s%g%s", "viento_capa_ini_esc_", h[h_idx],
				".inp");
		fichero = (char *)malloc((len + 1)* sizeof(char));
		snprintf(fichero, len+1, "%s%g%s", "viento_capa_ini_esc_", h[h_idx], ".inp");

		fp=fopen(fichero, "w");
		if (!fp) {
			fprintf(stderr, "Problema al intentar abrir el archivo %s", fichero);
            exit(EXIT_FAILURE);
		}

        fprintf(fp, "%zu %zu 1 0 0\n", terreno_capa.nx * terreno_capa.ny, ((terreno_capa.nx-1) * (terreno_capa.ny-1) * 2));

		for (size_t i=0; i<terreno_capa.nx; i++) {
			for (size_t j=0; j<terreno_capa.ny; j++) {
				size_t idx_terr = j+terreno_capa.nx*i;
                fprintf(fp, " %zu %f %f %f\n", idx_terr+1,
						terreno_capa.nodos[idx_terr][0],
						terreno_capa.nodos[idx_terr][1],
						terreno_capa.nodos[idx_terr][2]+h[h_idx]);
			}
		}

		{size_t k = 1;
			for (size_t i=0; i < (terreno_capa.nx-1) ; i++) {
				for (size_t j=0; j < (terreno_capa.ny-1); j++) {
                    fprintf (fp, "%zu %d %s %zu %zu %zu\n",
							k, 0, " tri ",
							(j*terreno_capa.nx)+i+1,
							(j*terreno_capa.nx)+i+2,((j+1)*terreno_capa.nx)+i+2);
					k++;
                    fprintf (fp, "%zu %d %s %zu %zu %zu\n",
							k, 0, " tri ",
							(j*terreno_capa.nx)+i+1,
							((j+1)*terreno_capa.nx)+i+2,((j+1)*terreno_capa.nx)+i+1);
					k++;
				}
			}
		}

		fprintf(fp, "1 1\nviento, ms-1\n");

		for (size_t i=0; i<terreno_capa.nx; i++) {
			for (size_t j=0; j<terreno_capa.ny; j++) {
				size_t idx_terr = j+terreno_capa.nx*i;
            fprintf (fp, "%zu %f \n", idx_terr+1, sqrt((capa_viento.wind[idx_terr].x*capa_viento.wind[idx_terr].x)+
					(capa_viento.wind[idx_terr].y*capa_viento.wind[idx_terr].y)+
					(capa_viento.wind[idx_terr].z*capa_viento.wind[idx_terr].z)));
			}
		}

		iostat = fclose(fp);
		if (iostat != 0) {
			fprintf(stderr, "Problema al cerrar el archivo %s", fichero);
            exit(EXIT_FAILURE);
		}

		free(fichero);

		free(capa_viento.wind);
		free(capa_viento_ini.wind);
	}


	for (size_t i=0; i < terreno_capa.nx*terreno_capa.ny; i++) {
		free(terreno_capa.nodos[i]);
	}

	free(h);
	free(terreno_capa.nodos);
}


