#ifndef CAPAS_H
#define CAPAS_H

#include "lectura.h"

/**
 * @file capas.h
 * @brief Módulo para crear los campos de viento en capas a distintas alturas
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @brief Estructura para guardar el campo de viento de la capa
 */
struct sVientoCapa {
    /**
     * @brief Vector de viento 3D
     */
    struct T_Vector3D *wind;
};

/**
 * @brief Estructura básica de la capa
 */
struct sCapaTerreno {
    /**
     * @brief Número de intervalos en la coordenada X
     */
    size_t nx;
    /**
     * @brief Número de intervalos en la coordenada Y
     */
    size_t ny;
    /**
     * @brief Puntero a las coordenadas de los nodos que conforman el la capa
     */
    double **nodos;
};

/**
 * @brief Función que calcula el campo de viento resultante en varias capas a distintas alturas, y lo escribe en un fichero
 * @param[in] malla Malla de referencia de donde se encuentran los valores de la velocidad interpolada
 *        y la velocidad resultante
 * @param[in] Datmos Estructura de datos atmosféricos para encontrar el viento interpolado en los puntos
 *        de las capas
 * @param[in] topo Valores de topografia para encontrar el viento interpolado en los puntos de las capas
 * @param[in] todas_est Estaciones de viento para encontrar el viento interpolado en los puntos
 *        de las capas
 * @param[in] file_format Tipo de origen de viento, harmonie o estaciones
 */

void calculaCapas(const struct T_Malla *malla,
                  const struct T_DatosAtmos *Datmos,
                  const struct T_Topografia *topo,
                  const struct T_NEstacion *todas_est,
                  const char *file_format);

#endif //CAPAS_H
