#include "caras_ext.h"

#include "comparisons.h"
#include "debugInfo.h"

#include <assert.h>
#include <stdlib.h>

/**
 * @file caras_ext.c
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

////////////////////////////////////////////////////////////////////////////////
//BuscarCaraInterior

void BuscarCaraInterior (const size_t tet, const size_t cara, const size_t *nods_cara, const size_t tet2,
                         const size_t cara2, const size_t *nods_cara2, size_t **tetraedros_caras)
{

    if (!tetraedros_caras[tet][cara] &&
            !tetraedros_caras[tet2][cara2]) {
        if((nods_cara[0] == nods_cara2[0] ||
            nods_cara[0] == nods_cara2[1] ||
            nods_cara[0] == nods_cara2[2]) &&
                (nods_cara[1] == nods_cara2[0] ||
                 nods_cara[1] == nods_cara2[1] ||
                 nods_cara[1] == nods_cara2[2]) &&
                (nods_cara[2] == nods_cara2[0] ||
                 nods_cara[2] == nods_cara2[1] ||
                 nods_cara[2] == nods_cara2[2])) {
            tetraedros_caras[tet][cara] = 1;
            tetraedros_caras[tet2][cara2] = 1;
        }
    }
}

////////////////////////////////////////////////
//DevuelveCaras

void DevuelveCaras (size_t no_cara, size_t *cara1_id, size_t *cara2_id, size_t *cara3_id)
{
    switch (no_cara) {
    case 1:
        *cara1_id = 0;
        *cara2_id = 1;
        *cara3_id = 2;
        break;
    case 2:
        *cara1_id = 0;
        *cara2_id = 1;
        *cara3_id = 3;
        break;
    case 3:
        *cara1_id = 0;
        *cara2_id = 2;
        *cara3_id = 3;
        break;
    case 4:
        *cara1_id = 1;
        *cara2_id = 2;
        *cara3_id = 3;
        break;
    }
}

void DevuelveNodosCara(const size_t **tetra, const size_t tet, const size_t cara, size_t *nodos_cara)
{
    switch (cara) {
    case 0:
        nodos_cara[0] = tetra[tet][0];
        nodos_cara[1] = tetra[tet][2];
        nodos_cara[2] = tetra[tet][1];
        break;
    case 1:
        nodos_cara[0] = tetra[tet][0];
        nodos_cara[1] = tetra[tet][1];
        nodos_cara[2] = tetra[tet][3];
        break;
    case 2:
        nodos_cara[0] = tetra[tet][0];
        nodos_cara[1] = tetra[tet][3];
        nodos_cara[2] = tetra[tet][2];
        break;
    case 3:
        nodos_cara[0] = tetra[tet][1];
        nodos_cara[1] = tetra[tet][2];
        nodos_cara[2] = tetra[tet][3];
        break;
    }
}

////////////////////////////////////////////////////////////////////////////////
//Any

bool any(const size_t scalar, const size_t* vector, const size_t size)
{
    for (size_t i=0; i < size; i++) {
        if (scalar == vector[i]) {
            return true;
        }
    }
    return false;
}

////////////////////////////////////////////////////////////////////////////////
//LocalizarCara

void LocalizarCara(const size_t tet, const size_t nodo1, const size_t nodo2, const size_t nodo3, const size_t ref1,
                   const size_t ref2, const size_t ref3, const size_t ref_cara, const size_t* cara, size_t **nodos_pared,
                   size_t *ref_caras)
{
    if(any(ref1, cara, 9)
            && any(ref2, cara, 9)
            && any(ref3, cara, 9)) {
        ref_caras[tet] = ref_cara;
        nodos_pared[nodo1][ref_cara-1] = 1;
        nodos_pared[nodo2][ref_cara-1] = 1;
        nodos_pared[nodo3][ref_cara-1] = 1;
    }
}

void CorreccionNumerosReferencia(struct T_Malla *malla)
{
    size_t ***tetraedros_nodos;
    size_t **tetraedros_caras;
    size_t *counter;
    size_t no_tetra_nodos=500;

    tetraedros_nodos = (size_t ***) malloc(malla->Num_nodos * sizeof(size_t **));
    if (tetraedros_nodos == NULL) {
        fprintf(stderr, "no se ha podido alocatar tetraedros_nodos\n");
        exit(EXIT_FAILURE);
    }
    tetraedros_caras = (size_t **) malloc(malla->Num_tetra * sizeof(size_t *));
    if (tetraedros_caras == NULL) {
        fprintf(stderr, "no se ha podido alocatar tetraedros_caras\n");
        exit(EXIT_FAILURE);
    }
    for (size_t i=0; i< malla->Num_nodos; i++) {
        tetraedros_nodos[i] = (size_t **) malloc(no_tetra_nodos * sizeof(size_t *));
        if (tetraedros_nodos[i] == NULL) {
            fprintf(stderr, "no se ha podido alocatar tetraedros_nodos[%zu]\n", i);
            exit(EXIT_FAILURE);
        }
        for (size_t j=0; j < no_tetra_nodos; j++) {
            tetraedros_nodos[i][j] = (size_t *) malloc(2 * sizeof(size_t));
            if (tetraedros_nodos[i][j] == NULL) {
                fprintf(stderr, "no se ha podido alocatar tetraedros_nodos[%zu][%zu]\n", i, j);
                exit(EXIT_FAILURE);
            }
        }
    }

    for (size_t i=0; i< malla->Num_tetra; i++) {
        tetraedros_caras[i] = (size_t *) malloc(4 * sizeof(size_t));
        if (tetraedros_caras[i] == NULL) {
            fprintf(stderr, "no se ha podido alocatar tetraedros_caras[%zu]\n", i);
            exit(EXIT_FAILURE);
        }
        tetraedros_caras[i][0] = 0;
        tetraedros_caras[i][1] = 0;
        tetraedros_caras[i][2] = 0;
        tetraedros_caras[i][3] = 0;
    }

    counter = (size_t *) malloc(malla->Num_nodos * sizeof(size_t));
    if (counter == NULL) {
        fprintf(stderr, "no se ha podido alocatar counter\n");
        exit(EXIT_FAILURE);
    }
    for (size_t i=0; i<malla->Num_nodos; i++) {
        counter[i]=0;
    }

    for (size_t tet=0; tet<malla->Num_tetra; tet++) {
        size_t nodo1 = malla->tetra[tet][0];
        size_t nodo2 = malla->tetra[tet][1];
        size_t nodo3 = malla->tetra[tet][2];
        size_t nodo4 = malla->tetra[tet][3];

        tetraedros_nodos[nodo1][counter[nodo1]][0]=tet;
        tetraedros_nodos[nodo1][counter[nodo1]][1]=1;
        tetraedros_nodos[nodo2][counter[nodo2]][0]=tet;
        tetraedros_nodos[nodo2][counter[nodo2]][1]=2;
        tetraedros_nodos[nodo3][counter[nodo3]][0]=tet;
        tetraedros_nodos[nodo3][counter[nodo3]][1]=3;
        tetraedros_nodos[nodo4][counter[nodo4]][0]=tet;
        tetraedros_nodos[nodo4][counter[nodo4]][1]=4;

        counter[nodo1]=counter[nodo1]+1;
        if (counter[nodo1] >= no_tetra_nodos) {
            fprintf(stderr, "El nodo %zu tiene mas elementos compartidos que "
                    "alocatado: %zu\n", nodo1, counter[nodo1]);
            exit(EXIT_FAILURE);
        }
        counter[nodo2]=counter[nodo2]+1;
        if (counter[nodo2] >= no_tetra_nodos) {
            fprintf(stderr, "El nodo %zu tiene mas elementos compartidos que "
                    "alocatado: %zu\n", nodo2, counter[nodo2]);
            exit(EXIT_FAILURE);
        }
        counter[nodo3]=counter[nodo3]+1;
        if (counter[nodo3] >= no_tetra_nodos) {
            fprintf(stderr, "El nodo %zu tiene mas elementos compartidos que "
                    "alocatado: %zu\n", nodo3, counter[nodo3]);
            exit(EXIT_FAILURE);
        }
        counter[nodo4]=counter[nodo4]+1;
        if (counter[nodo4] >= no_tetra_nodos) {
            fprintf(stderr, "El nodo %zu tiene mas elementos compartidos que "
                    "alocatado: %zu\n", nodo4, counter[nodo4]);
            exit(EXIT_FAILURE);
        }
    }

    for (size_t nodo = 0; nodo < malla->Num_nodos; nodo++) {
        for (size_t tet=0; tet < counter[nodo]; tet++) {
            size_t tetra = tetraedros_nodos[nodo][tet][0];
            size_t cara1;
            size_t cara2;
            size_t cara3;
            size_t nodos_cara1[3];
            size_t nodos_cara2[3];
            size_t nodos_cara3[3];
            DevuelveCaras(tetraedros_nodos[nodo][tet][1], &cara1, &cara2,
                          &cara3);
            if (tetraedros_caras[tetra][cara1] &&
                    tetraedros_caras[tetra][cara2] &&
                    tetraedros_caras[tetra][cara3]) {
                continue;
            }
            DevuelveNodosCara((const size_t **)malla->tetra, tetra, cara1, nodos_cara1);
            DevuelveNodosCara((const size_t **)malla->tetra, tetra, cara2, nodos_cara2);
            DevuelveNodosCara((const size_t **)malla->tetra, tetra, cara3, nodos_cara3);


            for (size_t tet_comp=tet+1; tet_comp < counter[nodo];
                 tet_comp++) {
                size_t tetra2 = tetraedros_nodos[nodo][tet_comp][0];
                size_t cara1_comp;
                size_t cara2_comp;
                size_t cara3_comp;
                size_t nodos_cara1_comp[3];
                size_t nodos_cara2_comp[3];
                size_t nodos_cara3_comp[3];

                DevuelveCaras(tetraedros_nodos[nodo][tet_comp][1],
                              &cara1_comp, &cara2_comp,
                              &cara3_comp);
                if (tetraedros_caras[tetra2][cara1_comp] &&
                        tetraedros_caras[tetra2][cara2_comp] &&
                        tetraedros_caras[tetra2][cara3_comp]) {
                    continue;
                }
                DevuelveNodosCara((const size_t **)malla->tetra, tetra2, cara1_comp, nodos_cara1_comp);
                DevuelveNodosCara((const size_t **)malla->tetra, tetra2, cara2_comp, nodos_cara2_comp);
                DevuelveNodosCara((const size_t **)malla->tetra, tetra2, cara3_comp, nodos_cara3_comp);

                BuscarCaraInterior(tetra, cara1, nodos_cara1,
                                   tetra2, cara1_comp,
                                   nodos_cara1_comp,
                                   tetraedros_caras);
                BuscarCaraInterior(tetra, cara1, nodos_cara1,
                                   tetra2, cara2_comp,
                                   nodos_cara2_comp,
                                   tetraedros_caras);
                BuscarCaraInterior(tetra, cara1, nodos_cara1,
                                   tetra2, cara3_comp,
                                   nodos_cara3_comp,
                                   tetraedros_caras);

                BuscarCaraInterior(tetra, cara2, nodos_cara2,
                                   tetra2, cara1_comp,
                                   nodos_cara1_comp,
                                   tetraedros_caras);
                BuscarCaraInterior(tetra, cara2, nodos_cara2,
                                   tetra2, cara2_comp,
                                   nodos_cara2_comp,
                                   tetraedros_caras);
                BuscarCaraInterior(tetra, cara2, nodos_cara2,
                                   tetra2, cara3_comp,
                                   nodos_cara3_comp,
                                   tetraedros_caras);

                BuscarCaraInterior(tetra, cara3, nodos_cara3,
                                   tetra2, cara1_comp,
                                   nodos_cara1_comp,
                                   tetraedros_caras);
                BuscarCaraInterior(tetra, cara3, nodos_cara3,
                                   tetra2, cara2_comp,
                                   nodos_cara2_comp,
                                   tetraedros_caras);
                BuscarCaraInterior(tetra, cara3, nodos_cara3,
                                   tetra2, cara3_comp,
                                   nodos_cara3_comp,
                                   tetraedros_caras);

                if (tetraedros_caras[tetra][cara1] &&
                        tetraedros_caras[tetra][cara2] &&
                        tetraedros_caras[tetra][cara3]) {
                    break;
                }
            }
        }
    }

    size_t no_caras_ext=0;
    size_t * nodos_caras_exteriores=(size_t *) malloc(malla->Num_nodos*sizeof(size_t));
    if (nodos_caras_exteriores == NULL) {
        fprintf(stderr, "no se ha podido alocatar nodos_caras_exteriores\n");
        exit(EXIT_FAILURE);
    }
    for (size_t nodo=0; nodo < malla->Num_nodos; nodo++) {
        nodos_caras_exteriores[nodo]=0;
    }

    for (size_t tet=0; tet < malla->Num_tetra; tet++) {
        for (size_t cara=0; cara < 4; cara++) {
            if(!tetraedros_caras[tet][cara]){
                no_caras_ext = no_caras_ext + 1;
                size_t nodos_cara[3];
                DevuelveNodosCara((const size_t **)malla->tetra, tet, cara, nodos_cara);
                nodos_caras_exteriores[nodos_cara[0]]=1;
                nodos_caras_exteriores[nodos_cara[1]]=1;
                nodos_caras_exteriores[nodos_cara[2]]=1;
            }
        }
    }

    size_t **caras_ext = (size_t **)malloc(no_caras_ext * sizeof(size_t *));
    if (caras_ext == NULL) {
        fprintf(stderr, "no se ha podido alocatar caras_ext\n");
        exit(EXIT_FAILURE);
    }

    no_caras_ext = 0;

    for (size_t tet=0; tet < malla->Num_tetra; tet++) {
        for (size_t cara=0; cara < 4; cara++) {
            if(!tetraedros_caras[tet][cara]){
                size_t nodos_cara[3];
                DevuelveNodosCara((const size_t **)malla->tetra, tet, cara, nodos_cara);
                caras_ext[no_caras_ext] = (size_t *)malloc(3*sizeof(size_t));
                if (caras_ext[no_caras_ext] == NULL) {
                    fprintf(stderr, "no se ha podido alocatar caras_ext[%zu]\n", no_caras_ext);
                    exit(EXIT_FAILURE);
                }
                caras_ext[no_caras_ext][0]= nodos_cara[0];
                caras_ext[no_caras_ext][1]= nodos_cara[1];
                caras_ext[no_caras_ext][2]= nodos_cara[2];
                no_caras_ext = no_caras_ext + 1;
            }
        }
    }

    size_t no_ext=0;
    size_t *nodos_ext=(size_t *)malloc(no_caras_ext * sizeof(size_t));
    if (nodos_ext == NULL) {
        fprintf(stderr, "no se ha podido alocatar nodos_ext\n");
        exit(EXIT_FAILURE);
    }
    for (size_t nodo=0; nodo < malla->Num_nodos; nodo++) {
        if (nodos_caras_exteriores[nodo]) {
            nodos_ext[no_ext] = nodo;
            no_ext = no_ext+1;
        }
    }

    double xmax=malla->nodo[0][0];
    double xmin=malla->nodo[0][0];
    double ymax=malla->nodo[0][1];
    double ymin=malla->nodo[0][1];
    double zmin=malla->nodo[0][2];
    double zmax=malla->nodo[0][2];

    for (size_t nodo=0; nodo < no_ext; nodo++) {
        double x=malla->nodo[nodos_ext[nodo]][0];
        double y=malla->nodo[nodos_ext[nodo]][1];
        double z=malla->nodo[nodos_ext[nodo]][2];

        xmin = fmin(xmin,x);
        ymin = fmin(ymin,y);
        zmin = fmin(zmin,z);

        xmax = fmax(xmax,x);
        ymax = fmax(ymax,y);
        zmax = fmax(zmax,z);
    }

    malla->terreno.x_max = xmax;
    malla->terreno.y_max = ymax;
    malla->terreno.x_min = xmin;
    malla->terreno.y_min = ymin;
    malla->terreno.z_min = zmin;
    malla->terreno.z_max = zmax;

    size_t no_1 = no_ext;
    size_t no_2 = no_ext;
    size_t no_3 = no_ext;
    size_t no_4 = no_ext;
    double z_no_1=zmax;
    double z_no_2=zmax;
    double z_no_3=zmax;
    double z_no_4=zmax;
    for (size_t nodo=0; nodo < no_ext; nodo++) {
        double x=malla->nodo[nodos_ext[nodo]][0];
        double y=malla->nodo[nodos_ext[nodo]][1];
        double z=malla->nodo[nodos_ext[nodo]][2];
        if (equal(&x,&xmin)) {
            if (equal(&y,&ymin)) {
                if (equal(&z, &zmax)) {
                    malla->condicion[nodos_ext[nodo]]=5;
                } else {
                    malla->condicion[nodos_ext[nodo]]=17;
                    if (lessThan(&z, &z_no_1)) {
                        z_no_1 = z;
                        no_1 = nodo;
                    }
                }
            } else if (equal(&y,&ymax)) {
                if (equal(&z,&zmax)) {
                    malla->condicion[nodos_ext[nodo]]=8;
                } else {
                    malla->condicion[nodos_ext[nodo]]=20;
                    if (lessThan(&z, &z_no_4)) {
                        z_no_4 = z;
                        no_4 = nodo;
                    }
                }
            } else {
                if (equal(&z, &zmax)) {
                    malla->condicion[nodos_ext[nodo]]=16;
                } else {
                    malla->condicion[nodos_ext[nodo]]=26;
                }
            }
        } else if (equal(&x, &xmax)) {
            if (equal(&y, &ymin)) {
                if (equal(&z,&zmax)) {
                    malla->condicion[nodos_ext[nodo]]=6;
                } else {
                    malla->condicion[nodos_ext[nodo]]=18;
                    if (lessThan(&z, &z_no_2)) {
                        z_no_2 = z;
                        no_2 = nodo;
                    }
                }
            } else if (equal(&y, &ymax)) {
                if (equal(&z, &zmax)) {
                    malla->condicion[nodos_ext[nodo]]=7;
                } else {
                    malla->condicion[nodos_ext[nodo]]=19;
                    if (lessThan(&z, &z_no_3)) {
                        z_no_3 = z;
                        no_3 = nodo;
                    }
                }
            } else {
                if (equal(&z, &zmax)) {
                    malla->condicion[nodos_ext[nodo]]=14;
                } else {
                    malla->condicion[nodos_ext[nodo]]=24;
                }
            }
        } else {
            if (equal(&y, &ymin)) {
                if (equal(&z, &zmax)) {
                    malla->condicion[nodos_ext[nodo]]=13;
                } else {
                    malla->condicion[nodos_ext[nodo]]=23;
                }
            } else if (equal(&y, &ymax)) {
                if (equal(&z, &zmax)) {
                    malla->condicion[nodos_ext[nodo]]=15;
                } else {
                    malla->condicion[nodos_ext[nodo]]=25;
                }
            } else {
                if (equal(&z, &zmax)) {
                    malla->condicion[nodos_ext[nodo]]=22;
                } else {
                    malla->condicion[nodos_ext[nodo]]=21;
                }
            }
        }
    }

    if (no_1 == no_ext){
        fprintf(stderr, "CorreccionNumeroReferencia: "
                "No se ha encontrado el nodo perteneciente al primer vértice del paralalepipedo\n");
        exit(EXIT_FAILURE);
    } else if (no_2 == no_ext){
        fprintf(stderr, "CorreccionNumeroReferencia: "
                "No se ha encontrado el nodo perteneciente al segundo vértice del paralalepipedo\n");
        exit(EXIT_FAILURE);
    } else if (no_3 == no_ext){
        fprintf(stderr, "CorreccionNumeroReferencia: "
                "No se ha encontrado el nodo perteneciente al tercer vértice del paralalepipedo\n");
        exit(EXIT_FAILURE);
    } else if (no_4 == no_ext){
        fprintf(stderr, "CorreccionNumeroReferencia: "
                "No se ha encontrado el nodo perteneciente al cuarto vértice del paralalepipedo\n");
        exit(EXIT_FAILURE);
    }

    malla->condicion[no_1] = 1;
    malla->condicion[no_2] = 2;
    malla->condicion[no_3] = 3;
    malla->condicion[no_4] = 4;

    size_t cara1[9] = {1, 2, 3, 4, 9, 10, 11, 12, 21};
    size_t cara2[9] = {5, 6, 7, 8, 13, 14, 15, 16, 22};
    size_t cara3[9] = {1, 2, 5, 6, 9, 13, 17, 18, 23};
    size_t cara4[9] = {2, 3, 6, 7, 10, 14, 18, 19, 24};
    size_t cara5[9] = {3, 4, 7, 8, 11, 15, 19, 20, 25};
    size_t cara6[9] = {1, 4, 5, 8, 12, 16, 17, 20, 26};

    size_t *ref_caras = (size_t *)malloc(no_caras_ext * sizeof(size_t));
    if (ref_caras == NULL) {
        fprintf(stderr, "no se ha podido alocatar ref_caras\n");
        exit(EXIT_FAILURE);
    }
    size_t **nodos_pared = (size_t **)malloc(malla->Num_nodos * sizeof(size_t*));
    if (nodos_pared == NULL) {
        fprintf(stderr, "no se ha podido alocatar nodos_pared\n");
        exit(EXIT_FAILURE);
    }
    for (size_t i=0; i<malla->Num_nodos; i++) {
        nodos_pared[i] = (size_t *)malloc(6 * sizeof(size_t));
        if (nodos_pared[i] == NULL) {
            fprintf(stderr, "no se ha podido alocatar nodos_pared[%zu]\n", i);
            exit(EXIT_FAILURE);
        }
        nodos_pared[i][0] = 0;
        nodos_pared[i][1] = 0;
        nodos_pared[i][2] = 0;
        nodos_pared[i][3] = 0;
        nodos_pared[i][4] = 0;
        nodos_pared[i][5] = 0;
    }


    //Para encontrar los nodos que pertenecen a las aristas inferiores
    //clasificamos caras
    malla->terreno.Num_caras = 0;
    for (size_t tet=0; tet < no_caras_ext; tet++) {
        size_t nodo1 = caras_ext[tet][0];
        size_t nodo2 = caras_ext[tet][1];
        size_t nodo3 = caras_ext[tet][2];
        size_t ref1 = malla->condicion[nodo1];
        size_t ref2 = malla->condicion[nodo2];
        size_t ref3 = malla->condicion[nodo3];

        ref_caras[tet] = 0;

        LocalizarCara(tet, nodo1, nodo2, nodo3, ref1, ref2, ref3, 1,
                      cara1, nodos_pared, ref_caras);
        LocalizarCara(tet, nodo1, nodo2, nodo3, ref1, ref2, ref3, 2,
                      cara2, nodos_pared, ref_caras);
        LocalizarCara(tet, nodo1, nodo2, nodo3, ref1, ref2, ref3, 3,
                      cara3, nodos_pared, ref_caras);
        LocalizarCara(tet, nodo1, nodo2, nodo3, ref1, ref2, ref3, 4,
                      cara4, nodos_pared, ref_caras);
        LocalizarCara(tet, nodo1, nodo2, nodo3, ref1, ref2, ref3, 5,
                      cara5, nodos_pared, ref_caras);
        LocalizarCara(tet, nodo1, nodo2, nodo3, ref1, ref2, ref3, 6,
                      cara6, nodos_pared, ref_caras);

        if (ref_caras[tet] == 0) {
            size_t carasuelo[17] = {1, 2, 3, 4, 9, 10, 11, 12, 17, 18, 19, 20, 23, 24, 25, 26, 21};
            if(any(ref1, carasuelo, 17)
                    && any(ref2, carasuelo, 17)
                    && any(ref3, carasuelo, 17)) {
                ref_caras[tet] = 1;
                for (size_t i=0; i<3; i++) {
                    size_t nodo;
                    size_t ref=0;
                    if (i == 0) {
                        nodo = nodo1;
                        ref = ref1;
                    } else if (i == 1) {
                        nodo = nodo2;
                        ref = ref2;
                    } else if (i == 2) {
                        nodo = nodo3;
                        ref = ref3;
                    }
                    if (ref == 23) {
                        malla->condicion[nodo] = 9;
                    } else if (ref == 24) {
                        malla->condicion[nodo] = 10;
                    } else if (ref == 25) {
                        malla->condicion[nodo] = 11;
                    } else if (ref == 26) {
                        malla->condicion[nodo] = 12;
                    }
                }
            } else {
                fprintf(stderr, "Elemento no localizado\n");
                fprintf(stderr, "Condicion: %zu %zu %zu\n", ref1, ref2, ref3);
                exit(EXIT_FAILURE);
            }
        }
        if (ref_caras[tet] == 1) {
            malla->terreno.Num_caras = malla->terreno.Num_caras + 1;
        }
    }

    malla->terreno.cara = (size_t **)malloc(malla->terreno.Num_caras * sizeof(size_t *));

    size_t no_caras_terreno =0;
    for (size_t tet=0; tet < no_caras_ext; tet++) {
        if (ref_caras[tet] == 1) {
            size_t nodo1 = caras_ext[tet][0];
            size_t nodo2 = caras_ext[tet][1];
            size_t nodo3 = caras_ext[tet][2];
            malla->terreno.cara[no_caras_terreno] = (size_t *)malloc(3 * sizeof(size_t));
            malla->terreno.cara[no_caras_terreno][0] = nodo1;
            malla->terreno.cara[no_caras_terreno][1] = nodo2;
            malla->terreno.cara[no_caras_terreno][2] = nodo3;
            no_caras_terreno = no_caras_terreno + 1;
        }
    }

    assert(no_caras_terreno == malla->terreno.Num_caras);

    for (size_t i = 0; i < malla->Num_nodos; i++) {
        for (size_t j = 0; j < no_tetra_nodos; j++){
            free(tetraedros_nodos[i][j]);
        }
        free(tetraedros_nodos[i]);
    }
    free(tetraedros_nodos);

    for (size_t i = 0; i < malla->Num_tetra; i++){
        free(tetraedros_caras[i]);
    }
    free(tetraedros_caras);

    free(counter);

    free(nodos_caras_exteriores);

    for (size_t i = 0; i < no_caras_ext; i++){
        free(caras_ext[i]);
    }
    free(caras_ext);

    free(nodos_ext);
    free(ref_caras);

    for (size_t i = 0; i < malla->Num_nodos; i++) {
        free(nodos_pared[i]);
    }
    free(nodos_pared);

}
