#include "ensamblaje.h"

#include "debugInfo.h"

#include <float.h>
#include <stdlib.h>

/**
 * @file ensamblaje.c
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

struct T_Vector_Nodos * CrearVectorDeListas(const size_t n)
{
    struct T_Vector_Nodos *matriz;

    // Reserva espacio para el vector de listas de ensamblaje
    matriz = (struct T_Vector_Nodos *)malloc(n * sizeof(struct T_Vector_Nodos));
    if (matriz == NULL) {
        fprintf(stderr, "CrearVectorDeListas: Imposible reservar memoria\n");
        exit(EXIT_FAILURE);
    }
    // inicializa el vector
    for(size_t i=0; i < n; i++) {
        matriz[i].count = 0;
        matriz[i].ptr = NULL;
    }
    return(matriz);
}

void
LiberarVectorDeListas(struct T_Vector_Nodos *matriz, const size_t n)
{
    struct T_Nodo *p1, *p2;
    for(size_t i = 0; i < n; i++){
        // libera los nodos de la lista encadenada
        p1 = p2 = matriz[i].ptr;
        while (p1 != NULL){
            p2 = p1->ptr;
            free(p1);
            p1 = p2;
        }
    }
    // libera el vector de listas
    free(matriz);
    matriz = NULL;
}

struct 	T_MElemental *
        CrearMatrizElemental(const size_t dim)
{
    struct T_MElemental *eltal1;

    if ((eltal1 = (struct T_MElemental *)malloc(sizeof(struct T_MElemental))) == NULL){
        fprintf(stderr,"CrearMatrizElemental: No se ha podido alocatar eltarl1\n");
        exit(EXIT_FAILURE);
    }

    eltal1->dim = dim;
    eltal1->indice = (int *)malloc(eltal1->dim * sizeof(int));

    if ((eltal1->valor =
         (double **)malloc(eltal1->dim * sizeof(double *))) == NULL){
        fprintf(stderr,"CrearMatrizElemental: No se ha podido alocatar eltarl1->valor\n");
        exit(EXIT_FAILURE);
    }
    for (size_t i=0; i < eltal1->dim; i++)
        if ((eltal1->valor[i] =
             (double *)malloc(eltal1->dim * sizeof(double))) == NULL){
            fprintf(stderr,"CrearMatrizElemental: No se ha podido alocatar eltarl1->valor[%zu]\n", i);
            exit(EXIT_FAILURE);
        }
    return(eltal1);
}

void
DestruirMatrizElemental(struct T_MElemental *eltal1)
{
    for (size_t i=0; i < eltal1->dim; i++) {
        free(eltal1->valor[i]);
    }

    free(eltal1->valor);
    free(eltal1->indice);
    free(eltal1);
} // DestruirMatrizElemental


void InsertarNodo(struct T_Vector_Nodos *Matriz, const size_t fila, const size_t col, const double valor,
                  const double xp)
{
    struct T_Nodo *p, *p1, *p2;

    if ( (p = (struct T_Nodo *)malloc(sizeof(struct T_Nodo))) == NULL ){
        fprintf(stderr, "InsertarNodo: No se ha podido alocatar p\n");
        exit(EXIT_FAILURE);
    }
    p->ptr = NULL;
    p->col = col;
    p->valor = valor;
    if (Matriz[fila].ptr == NULL){
        Matriz[fila].count = 1;
        Matriz[fila].ptr = p;
        Matriz[fila].xp = xp;
        return;
    }

    Matriz[fila].xp += xp;
    p1 = Matriz[fila].ptr;
    if (col < p1->col){
        // si va antes que el primero de la lista
        p ->ptr = p1;
        Matriz[fila].ptr = p;
        Matriz[fila].count++;
        return;
    }
    if (p1->col == col){
        // si es de la misma columna que el primero de la lista
        p1->valor += valor;
        free(p);
        return;
    }

    p2 = p1;
    while (p2->col < col) {
        p1 = p2;
        p2 = p2->ptr;
        if (p2 == NULL) break;
    }
    if (p2 == NULL){
        // inserta al final de la lista
        p1->ptr = p;
    } else {
        if (p2->col == col){
            p2->valor += valor;
            free(p);
            return;
        } else {
            p->ptr = p2;
            p1->ptr = p;
        }
    }
    Matriz[fila].count++;
    return;
}// InsertarNodo

int Condicion(const size_t ref)
{
    switch (ref){
    case 0:
        return(0);
        break;
    case 1:
        return(1);
        break;
    case 2:
        return(1);
        break;
    case 3:
        return(1);
        break;
    case 4:
        return(1);
        break;
    case 5:
        return(1);
        break;
    case 6:
        return(1);
        break;
    case 7:
        return(1);
        break;
    case 8:
        return(1);
        break;
    case 9:
        return(1);
        break;
    case 10:
        return(1);
        break;
    case 11:
        return(1);
        break;
    case 12:
        return(1);
        break;
    case 13:
        return(1);
        break;
    case 14:
        return(1);
        break;
    case 15:
        return(1);
        break;
    case 16:
        return(1);
        break;
    case 17:
        return(1);
        break;
    case 18:
        return(1);
        break;
    case 19:
        return(1);
        break;
    case 20:
        return(1);
        break;
    case 21:
        return(-1);
        break;
    case 22:
        return(-1);
        break;
    case 23:
        return(1);
        break;
    case 24:
        return(1);
        break;
    case 25:
        return(1);
        break;
    case 26:
        return(1);
        break;
    default:
        fprintf(stderr, "Condicion: Numero de referencia erroneo = %zu\n", ref);
        exit(EXIT_FAILURE);
        break;
    }
    // Si llega aqui es que el numero de referencia esta fuera del
    // rango [0,26], y por tanto es erroneo
    fprintf(stderr, "Condicion: Numero de referencia erroneo = %zu\n", ref);
    exit(EXIT_FAILURE);
}// Condicion

void
EnsamblarMatrizElemental(struct T_Vector_Nodos *Matriz, const struct T_MElemental *eltal,
                         const struct T_Malla *malla, const double *xp)
{

    for (size_t i=0; i < eltal->dim; i++){
        if (Condicion(malla->condicion[eltal->indice[i]]) > 0) continue;
        double x = xp[i];
        for (size_t j=0; j < eltal->dim; j++){
            if (Condicion(malla->condicion[eltal->indice[j]]) <= 0){
                InsertarNodo(Matriz, eltal->indice[i], eltal->indice[j],
                             eltal->valor[i][j], x);
                x = 0.0;
            }
        }
    }
}// EnsamblarMatrizElemental


/**
 * Sistema reducido: S.E. sin la ecuaciones correspondientes a
 * los terminos con condicion tipo Dirichlet.
 * Vector de reensamblaje: vector que contiene las posiciones de
 * los terminos tipo Dirichlet; se usa para construir la
 * solucion completa del sistema.
 */
size_t ConstruirMatrizMorse(const struct T_Vector_Nodos *Matriz, const size_t dim,
                            struct matrizmorse *morse, double **x, size_t **P)
{
    size_t  c=0;
    struct T_Nodo *p1, *p2;
    size_t *r = NULL;            // vector de reensamblaje


    morse->nonulos = 0;
    for (size_t i=0; i < dim; i++){
        morse->nonulos += Matriz[i].count;
        if (Matriz[i].ptr == NULL) c++;
    }
    morse->dimension = dim - c;
    // reserva memoria para el vector de reensamblaje
    if ((r = (size_t *)malloc((c+1) * sizeof(size_t))) == NULL){
        fprintf(stderr,"ConstruirMatrizMorse: No se ha podido alocatar r\n");
        exit(EXIT_FAILURE);
    }
    // Reserva memoria para el vector de t. indepen.
    if ((*x = (double *)malloc((dim - c) * sizeof(double))) == NULL){
        fprintf(stderr,"ConstruirMatrizMorse: No se ha podido alocatar x\n");
        exit(EXIT_FAILURE);
    }
    // Reserva memoria para la matriz morse
    if ( (morse->valor =
          (double *)malloc(morse->nonulos * sizeof(double))) == NULL){
        fprintf(stderr,"ConstruirMatrizMorse: No se ha podido alocatar morse->valor\n");
        exit(EXIT_FAILURE);
    }
    if ( (morse->columna =
          (size_t *)malloc(morse->nonulos * sizeof(size_t))) == NULL){
        fprintf(stderr,"ConstruirMatrizMorse: No se ha podido alocatar morse->columna\n");
        exit(EXIT_FAILURE);
    }
    if ( (morse->fila =
          (size_t *)malloc((dim + 1) * sizeof(size_t))) == NULL){
        fprintf(stderr,"ConstruirMatrizMorse: No se ha podido alocatar morse->fila\n");
        exit(EXIT_FAILURE);
    }

    {
        size_t fil = 0;
        morse->fila[fil] =  0;
        size_t val = 0;
        for (size_t i = c = 0; i < dim; i++){
            p1 = p2 = Matriz[i].ptr;
            if (p1 == NULL){
                r[c] = i;
                c++;
            } else {
                x[0][fil] = Matriz[i].xp;
                fil++;

                while (p1 != NULL){
                    morse->valor[val] = p1->valor;
                    morse->columna[val] = p1->col;
                    p2 = p1->ptr;
                    p1 = p2;
                    val++;
                }
                morse->fila[fil] = morse->fila[fil-1] + Matriz[i].count;
            }
        }
    }
    //Marca de final del vector. En la funcion reensamblar se buscara el
    //ultimo valor de r que no apunta a ninguna fila
    r[c] = dim+1;
    *P = r;

    // Ajusta las columnas: debido a la desaparicion de filas y
    // columnas completas pertenecientes a los nodos tipo
    // Dirichlet, hay que ajustar las columnas de la matriz morse
    for(size_t i=0; i < c-1; i++)
    {
        for(size_t j=0; j < morse->nonulos; j++)
        {
            if ((morse->columna[j] > r[i]) && (morse->columna[j] < r[i+1]))
            {
                morse->columna[j] -= (i+1);
            }
        }
    }

    {
        size_t aux = c - 1;
        for(size_t j=0; j < morse->nonulos; j++)
        {
            if (morse->columna[j] > r[aux])
            {
                morse->columna[j] -= c;
            }
        }
    }

    return(c);
}

double * Reensamblar(const size_t dim, const double *x, size_t *r)
{
    int p, t;
    double *sol;

    if ((sol = (double *)malloc(dim * sizeof(double))) == NULL){
        fprintf(stderr,"Reensamblar: No se ha podido alocatar sol\n");
        exit(EXIT_FAILURE);
    }

    t = p = 0;
    for(size_t i=0; i<dim; i++){
        if (i != r[p]){
            // rensamblar solucion obtenida
            sol[i] = x[t];
            t++;
        } else {
            // rensamblar condicion Dirichlet
            // OJO : buscar un mecanismo para invocar a la funcion
            // definida por el vector de condiciones de contorno
            // Algo del tipo:
            // sol[i] = funcion(malla->condicion[i]);
            sol[i] = 0.0;
            p++;
        }
    }
    // libera la memoria ocupada por los vectores solucion y
    // de reensamblaje
    // free(x);
    free(r);
    return(sol);
}

double *
ResolverSistema(struct T_Vector_Nodos *Matriz, const size_t dim, const int metodo,
                const int precond, const double w)
{
    size_t *P;             // vector de reensamblaje
    double *b;          // vector de T.I. reducido
    double *vini;       // solucion inicial para BiCGSTAB
    double *sol;        // solucion del sistema reducido
    struct matrizmorse *morse; // matriz del sistema reducido
    TFuncRemonte funcion_remonte; // fcion. remonte para BICGSTAB
    TFuncPrecond f_precond;       // fcion. de precondicionamiento
    TFuncMetodo solver; // metodo de resolucion: CG o BiCGSTAB
    double *W = NULL;   // vector precondicionador
    size_t *posdiag_morse; // vector de posiciones de la diagonal de la
    size_t *posdiag_W;     // idem del vector precondicionador


    DEBUG_INFO(
                FILE *f1;
            FILE *f2;
    );

    f_precond = NULL;
    if ((morse = (struct matrizmorse *)malloc(sizeof(struct matrizmorse)))
            == NULL) {
        fprintf(stderr,"ResolverSistema: No se ha podido alocatar morse\n");
        exit(EXIT_FAILURE);
    }

    ConstruirMatrizMorse(Matriz, dim, morse, &b, &P);

    switch (metodo){
    case 0: // Gradiente Conjugado
        solver = GradienteConjugado;
        break;
    case 1: // BiCGSTAB
        solver = BICGSTAB;
        break;
    default:
        fprintf(stderr,"ResolverSistema: No existe ese metodo de resolucion\n");
        exit(EXIT_FAILURE);
    }

    switch (precond){
    case 1: // Sin precondicionar
        f_precond = NULL;
        funcion_remonte = RemonteIdent;
        break;
    case 2: // Precondicionador diagonal
        f_precond = PrecondDiag;
        funcion_remonte = RemonteDiag;
        break;
    case 3: // Precondicionador ILU
        f_precond = PrecondILU;
        funcion_remonte = RemonteLU;
        break;
    case 4: // Precondicionador SSOR
        f_precond = PrecondSSOR_NS;
        funcion_remonte = RemonteLU;
        break;
    case 7: // Precondicionador diagonal optimo
        f_precond=PrecondDiagOptimo;
        funcion_remonte = RemonteDiag;
        break;
    default:
        fprintf(stderr,"ResolverSistema: No existe ese precondicionador\n");
        exit(EXIT_FAILURE);
    }

    // Crea solucion inicial y solucion
    vini = (double *)malloc(morse->dimension * sizeof(double));
    sol = (double *)malloc(morse->dimension * sizeof(double));
    if ((vini == NULL) || (sol == NULL)) {
        fprintf(stderr,"ResolverSistema: No se ha podido alocatar vini o sol\n");
        exit(EXIT_FAILURE);
    }
    for (size_t i=0; i < morse->dimension; i++) vini[i] = 0.0;
    // Obtiene las posiciones de la diagonal de la matriz del sistema
    if ((posdiag_morse = (size_t *)malloc(morse->dimension * sizeof(size_t)))
            == NULL) {
        fprintf(stderr,"ResolverSistema: No se ha podido alocatar posdiag_morse\n");
        exit(EXIT_FAILURE);
    }
    PosDiagonalMorse(morse, posdiag_morse);

    // Calcula el precondicionador y su posdiag (que es el de morse)
    if ((posdiag_W=(size_t*)malloc(morse->dimension*sizeof(size_t))) == NULL){
        fprintf(stderr,"ResolverSistema: No se ha podido alocatar posdiag_W\n");
        exit(EXIT_FAILURE);
    }
    if (f_precond != NULL)
    {

        W = f_precond(morse, posdiag_morse, &w, posdiag_W);

    }

    DEBUG_INFO(

                f1=fopen("00morse", "w"); f2=fopen("00diag", "w");
    if((f1==NULL)||(f2==NULL)){
        fprintf(stderr,"ResolverSistema: No se han podido abrir los ficheros 00morse 00diag\n");
        exit(EXIT_FAILURE);
    }
    for (size_t i=0; i<morse->nonulos; i++)
        fprintf(f1, "%10.6f\n", morse->valor[i]);
    for(size_t i=0; i<morse->dimension; i++)
        fprintf(f2, "%10.6f\n", morse->valor[posdiag_morse[i]]);
    fclose(f1); fclose(f2);
    f1=fopen("00b", "w");
    if(f1==NULL){
        fprintf(stderr,"ResolverSistema: No se ha podido abrir el fichero 00b\n");
        exit(EXIT_FAILURE);
    }
    for (size_t i=0; i<morse->dimension; i++)
        fprintf(f1, "%10.6f\n", b[i]);
    fclose(f1);

    );

    const double tolerancia = 1.0e-12;

    if (solver(morse, b, sol, vini, morse->dimension + 3,
               tolerancia, W,
               posdiag_W, funcion_remonte) == 1){
    } else {
        fprintf(stderr,"ResolverSistema: El sistema no tiene solucion\n");
        exit(EXIT_FAILURE);
    }

    // libera memoria
    free(posdiag_morse);
    if (W != NULL) free(W);
    free(posdiag_W);
    free(vini);
    free(b);
    DestruirMorse(morse);
    // Reensambla los terminos tipo Dirichlet
    vini = Reensamblar(dim, sol, P);
    free(sol);
    return(vini);
}// ResolverSistema
