#include "funciones.h"

#include "comparisons.h"
#include <math.h>

/**
 * @file funciones.c
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

double DistanciaR2(const double *x1, const double *y1, const double *x2, const double *y2)
{
    double x = (*x2) - (*x1);
    double y = (*y2) - (*y1);

    x = x*x;
    y = y*y;

    x = sqrt(x + y);
    return(x);
}

double DistanciaR3(const double *x1, const double *y1, const double *z1,
                   const double *x2, const double *y2, const double *z2)
{
    double x = (*x2) - (*x1);
    double y = (*y2) - (*y1);
    double z = (*z2) - (*z1);

    x = x*x;
    y = y*y;
    z = z*z;

    x = sqrt(x + y + z);
    return(x);
}

// Spline cubica usada en la interpolacion
double SplineCubica(const double *s)
{
    const double s2 = (*s) * (*s);
    const double s3 = s2 * (*s);
    const double s4 = s3 * (*s);

    const double one = 1.0;

    if (greatThan(s,&one)) {
        return (0.0);
    }

    double spline = 1 - 6*s2 + 8*s3 - 3*s4;
    return(spline);
}
