#ifndef  FUNCIONES_H
#define  FUNCIONES_H
/**
 * @file funciones.h
 * @brief Funciones auxiliares
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @brief Función que busca la distancia euclidea en 2D entre dos puntos
 * @param[in] x1,y1 Coordenada X e Y del punto 1
 * @param[in] x2,y2 Coordenada X e Y del punto 2
 * @return Distancia euclidea entre los dos puntos
 */
double DistanciaR2(const double *x1, const double *y1, const double *x2, const double *y2);

/**
 * @brief Función que busca la distancia euclidea en 3D entre dos puntos
 * @param[in] x1,y1,z1 Coordenada X Y Z del punto 1
 * @param[in] x2,y2,z2 Coordenada X Y Z del punto 2
 * @return Distancia euclidea entre los dos puntos
 */
double DistanciaR3(const double *x1, const double *y1, const double *z1,
           const double *x2, const double *y2, const double *z2);

/**
 * @brief Función que interpola según una spline cúbica
 * @param[in] s Valor de interpolación
 * @return Valor interpolado
 */
double SplineCubica(const double *s);


#endif //FUNCIONES_H
