#include "interpolacion.h"

#include "debugInfo.h"
#include "funciones.h"
#include "lectura.h"

#include <assert.h>
#include <float.h>
#include <math.h>
#include <stdlib.h>


/**
 * @file interpolacion.c
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


// Busca los incrementos maximos
void BuscarIncrementosHARM(struct VariableHARMONIE *q)
{
    double dx = -1;
    double dy = -1;
    double dxmin = DBL_MAX;
    double dymin = DBL_MAX;

    // Busca la minima diferencia de latitud entre dos filas
    // consecutivas del grid
    for (size_t j=0; j < q->dimx ; j++) {
        double y0, y1;
        for (size_t i=0; i < q->dimy - 1; i++) {
            //	  coordenadasUTMij(q, i  , j, t, &x0, &y0);
            //	  coordenadasUTMij(q, i+1, j, t, &x1, &y1);
            y0 = (double)q->latitud[i * q->dimx + j];
            y1 = (double)q->latitud[(i+1) * q->dimx + j];

            double disty = (double)fabs(y1 - y0);
            dy = (disty > dy) ? disty : dy ;
            dymin = (disty < dymin) ? disty : dymin ;
        }
    }

    for (size_t i = 0; i < q->dimy; i++) {
        double x0, x1;
        for (size_t j=0; j < q->dimx - 1; j++) {
            //	  coordenadasUTMij(q, i, j  , t, &x0, &y0);
            //	  coordenadasUTMij(q, i, j+1, t, &x1, &y1);
            x0 = (double)q->longitud[i*q->dimx + j];
            x1 = (double)q->longitud[i*q->dimx + j + 1];

            double distx = (double)fabs(x1 - x0);
            dx = (distx > dx) ? distx : dx ;
            dxmin = (distx < dxmin) ? distx : dxmin ;
        }
    }

    q->deltaxmax = dx;      q->deltaymax = dy;
    q->deltaxmin = dxmin;   q->deltaymin = dymin;

    DEBUG_INFO(
                fprintf(stdout, "BuscarIncrementosHAR: deltaxmax %f  deltaymax %f    deltaxmin %f deltaymin %f\n",
                       q->deltaxmax, q->deltaymax, q->deltaxmin, q->deltaymin);
            );
}


void BuscarBoundingBoxHARM(struct VariableHARMONIE *q)
{
    double xmax = -1;
    double ymax = -1;
    double xmin = FLT_MAX;
    double ymin = FLT_MAX;
    double xmaxmin = -1;
    double ymaxmin = -1;

    // Busco en la fila inferior
    for (size_t j = 0; j < q->dimx; j++) {
        size_t i = 0;
        double x = q->longitud[i*q->dimx + j];
        double y = q->latitud[i*q->dimx + j];
        //      coordenadasUTMij(q, i, j, t, &x, &y);
        if (x < xmin)  xmin = x;
        if (x > xmax)  xmax = x;
        if (y < ymin)  ymin = y;
        if (y > ymax)  ymax = y;
        if (y > ymaxmin) ymaxmin = y;
    }

    // Busco en la columna izquierda
    for (size_t i = 0; i < q->dimy; i++) {
        size_t j = 0;
        double x = q->longitud[i*q->dimx + j];
        double y = q->latitud[i*q->dimx + j];;
        //      coordenadasUTMij(q, i, j, t, &x, &y);
        if (x < xmin)  xmin = x;
        if (x > xmax)  xmax = x;
        if (y < ymin)  ymin = y;
        if (y > ymax)  ymax = y;
        if (x > xmaxmin) xmaxmin = x;
    }

    // Busco en la fila superior
    for (size_t j = 0; j < q->dimx; j++) {
        size_t i = q->dimy - 1;
        double x = q->longitud[i*q->dimx + j];
        double y = q->latitud[i*q->dimx + j];;
        //      coordenadasUTMij(q, i, j, t, &x, &y);
        if (x < xmin)  xmin = x;
        if (x > xmax)  xmax = x;
        if (y < ymin)  ymin = y;
        if (y > ymax)  ymax = y;
    }

    // Busco en la columna derecha
    for (size_t i = 0; i < q->dimy; i++) {
        size_t j = q->dimx - 1;
        double x = q->longitud[i*q->dimx + j];
        double y = q->latitud[i*q->dimx + j];
        //      coordenadasUTMij(q, i, j, t, &x, &y);
        if (x < xmin)  xmin = x;
        if (x > xmax)  xmax = x;
        if (y < ymin)  ymin = y;
        if (y > ymax)  ymax = y;
    }

    q->xmin = xmin;  q->ymin = ymin;
    q->xmax = xmax;  q->ymax = ymax;
    q->xmaxmin = xmaxmin;
    q->ymaxmin = ymaxmin;


    DEBUG_INFO(
                fprintf(stdout, "bounding box %f %f    %f %f\n", xmin, ymin, xmax, ymax);
            );
}

// Lee las tres componentes de viento (U, V, W) del fichero de HARMONIE
// Todo queda almacenado en una variable de tipo VariableHARMONIE
void LeerVientoHARMONIE(const char *nombreFichero, struct VariableHARMONIE *q)
{
    static char *nfun = "LeerVientoHARMONIE";
    FILE *f;
    size_t line_number = 0;
    char buffer[2048];

    // Abro el fichero
    if ( (f=fopen(nombreFichero, "r")) == NULL) {
        fprintf(stderr, "%s: Error al abrir el fichero %s\n",
                nfun, nombreFichero);
        exit(EXIT_FAILURE);
    }

    // Leo la primera linea del fichero
    int status = LeeLinea(f, buffer, sizeof(buffer), &line_number);
    if (status != 0)
    {
        fprintf(stderr, "%s: Error al leer las dimensiones\n", nfun);
        fprintf(stderr, "%s:%zu\n",nombreFichero, line_number);
        exit(EXIT_FAILURE);
    }
    {
        size_t dim[3];
        parseLineAsListOfUInts(3, buffer, nombreFichero, line_number, dim);

        q->dimx = dim[0];
        q->dimy = dim[1];
        q->dimz = dim[2];
    }


    //Reservo memoria
    size_t k = q->dimx * q->dimy * q->dimz;
    q->data_u = (double *)malloc(k*sizeof(double));
    q->data_v = (double *)malloc(k*sizeof(double));
    q->data_w = (double *)malloc(k*sizeof(double));
    q->altura = (double *)malloc(k*sizeof(double));
    if ((q->data_u == NULL) || (q->data_v == NULL) ||
            (q->data_w == NULL) || (q->altura == NULL) ) {
        fprintf(stderr, "%s: Error al reservar memoria para datos\n", nfun);
        exit(EXIT_FAILURE);
    }

    k = q->dimx * q->dimy;
    q->longitud = (double *)malloc(k*sizeof(double));
    q->latitud  = (double *)malloc(k*sizeof(double));
    if ((q->longitud == NULL) || (q->latitud == NULL)) {
        fprintf(stderr, "%s: Error al reservar memoria para coordenadas\n",
                nfun);
        exit(EXIT_FAILURE);
    }


    // Leo linea a linea y me asegura de que cada linea tiene 6 datos:
    // utmx, utmy, z, u, v, w
    {
        size_t i = 0;
        while (i < k){
            if (LeeLinea(f, buffer, sizeof(buffer), &line_number) == 0) {
                double a[6];
                parseLineAsListOfDoubles(6, buffer, nombreFichero, line_number, a);

                if (i < q->dimx * q->dimy) {
                    q->longitud[i] = a[0];
                    q->latitud[i]  = a[1];
                }
                q->altura[i] = a[2];
                q->data_u[i] = a[3];
                q->data_v[i] = a[4];
                q->data_w[i] = a[5];

                i++;
            } else {
                fprintf(stderr,"Problema leyendo los datos\n");
                fprintf(stderr, "%s:%zu\n",nombreFichero, line_number);
                exit(EXIT_FAILURE);
            }
        }


        // Se cierra el fichero
        fclose(f);

    }

    // Calcular las alturas y los incrementos usando el instante 0
    BuscarIncrementosHARM(q);
    BuscarBoundingBoxHARM(q);
}

// Devuelve la distancia entre el punto (utmx,utmy) y las correspondientes
// coordenadas utm la variable q en (i,j)
double DistanciaR2ijHARM(const struct VariableHARMONIE *q, const double* utmx, const double* utmy,
                         const int i, const int j)
{
    double ux = q->longitud[i*q->dimx + j];
    double uy = q->latitud[i*q->dimx + j];

    //  coordenadasUTMij(q, i, j, t, &ux, &uy);
    double d = DistanciaR2(utmx, utmy, &ux, &uy);
    return(d);
}

/**
 * Version con ventana de busqueda; busca desde el i0,j0
 * hasta el i0+ki,j0+kj . Esto no siempre ha de funcionar.
 * Depende de lo deformada que este la rejilla
 */
void LocalizarEnRejillaHARM(const double* utmx, const double* utmy, const double* utmz,
                            const struct VariableHARMONIE *q,
                            int *qi, int *qj, int *qk)
{
    size_t i, j, k;
    size_t iopt, jopt;
    size_t ki, kj;
    double dmin;
    double dist;

    size_t j0 = (size_t) trunc(((*utmx) - q->xmaxmin) / q->deltaxmax);
    size_t i0 = (size_t) trunc(((*utmy) - q->ymaxmin) / q->deltaymax);

    size_t jn = (size_t) trunc(((*utmx) - q->xmin) / q->deltaxmin) + 1;
    size_t in = (size_t) trunc(((*utmy) - q->ymin) / q->deltaymin) + 1;

    // Algunos sanity checks
    // Poner aqui algunas comprobaciones de i0, j0, in, jn
    assert(i0 < q->dimy);
    assert(j0 < q->dimx);


    DEBUG_INFO(
                fprintf(stdout, "LocalizarEnRejillaHARM: ventana de busqueda %zu %zu\n",
                       in-i0, jn-j0);
            );

    // Controla que se solicita un i,j que esta en el interior de
    // la rejilla
    assert(i0 < q->dimy);
    assert(j0 < q->dimx);

    *qi = *qj = *qk = -1;

    // Controla que el punto que se pide este en el bounding box
    // de la rejilla
    assert((*utmx >= q->xmin) && (*utmx <= q->xmax));
    assert((*utmy >= q->ymin) && (*utmy <= q->ymax));

    // Empiezo en i
    i = iopt = i0;    j = jopt = j0;


    DEBUG_INFO(
                fprintf(stdout, "LocalizarEnRejillaHARM:punto tentativo (%zu, %zu)\n", i, j);
            );

    dmin = FLT_MAX;

    // Busca en una ventana de -2 a +2 a partir de i0,j0
    for (ki=0; ki <= in-i0; ki++) {
        if (i+ki >= q->dimy) continue;
        for (kj=0; kj <= jn-j0; kj++) {
            if (j+kj >= q->dimx) continue;
            dist = DistanciaR2ijHARM(q, utmx, utmy, i+ki, j+kj);
            if (dist < dmin)
            {
                dmin = dist;
                iopt = i+ki;
                jopt = j+kj;
            }
        }
    }
    *qi = iopt;
    *qj = jopt;

    // Ahora hay que buscar en altura
    dmin = 1.0e100;
    *qk = -1;
    for (k=0; k < q->dimz; k++)
    {
        //      double z = obtener_altura(q, *qi, *qj, k, t);
        double z = q->altura[(k*q->dimx*q->dimy) + (iopt * q->dimx + jopt)];
        dist = fabs((*utmz) - z);
        if (dist < dmin)
        {
            dmin = dist;
            *qk = k;
        }
    }

    DEBUG_INFO(
    {
                    double X = q->longitud[iopt*q->dimx + jopt];
                    double Y = q->latitud[iopt*q->dimx + jopt];
                    double Z = q->altura[((*qk)*q->dimx*q->dimy) + (iopt * q->dimx + jopt)];
                    fprintf(stdout, "LocalizarEnRejillaHARM %f %f %f  ---> (%d %d %d)  %f %f %f\n",
                    *utmx, *utmy, *utmz, *qi, *qj, *qk,
                    X, Y, Z);
                }
                );


}




// Dados los indices i, j, k devuelve las coordenas UTM correspondientes
// de la variable *q
// i para la Y (latitud)
// j para la X (longitud)
// k para la altura
void Utm3DHARM(const struct VariableHARMONIE *q, size_t i, size_t j, size_t k,
               double *utmx, double *utmy, double *utmz)
{
    double x, y, z;

    //  coordenadasUTMij(q, i, j, t, &x, &y );
    //  z = obtener_altura(q, i, j, k, t);
    assert((k<q->dimz));
    int offset_xy = i*q->dimx + j;
    int offset_xyz = k * q->dimy * q->dimx + offset_xy;


    x = q->longitud[offset_xy];
    y = q->latitud[offset_xy];
    z = q->altura[offset_xyz];

    *utmx = x;
    *utmy = y;
    *utmz = z;

    DEBUG_INFO(
                fprintf(stdout, "Utm3DHARM %zu %zu %zu     %f %f %f\n",
                       i, j, k, x, y, z);
            );

}


void InterpolaComponente7VecinosHARM(const struct VariableHARMONIE *q, const double* x, const double* y,
                                     const double* z, double wind[3])
{
    int qi, qj, qk;
    size_t idx_vecino[7][3];
    double utm[7][3];

    double cte = 1.5; // Cte. que controla el radio de influencia de la funcion
    // interpoladora.Rango de cte: 1.0 <= cte < inf
    int i, k;

    LocalizarEnRejillaHARM(x, y, z, q, &qi, &qj, &qk);
    k = 0;
    idx_vecino[k][0] = qi;
    idx_vecino[k][1] = qj;
    idx_vecino[k][2] = qk;
    Utm3DHARM(q, idx_vecino[k][0], idx_vecino[k][1], idx_vecino[k][2],
            &utm[k][0], &utm[k][1], &utm[k][2]);
    k++;
    int dimx = q->dimx;
    int dimy = q->dimy;
    int dimz = q->dimz;
    // Compruebo vecino i+1, j, z
    if ((qi+1) < dimy) {
        idx_vecino[k][0] = qi+1;
        idx_vecino[k][1] = qj;
        idx_vecino[k][2] = qk;
        Utm3DHARM(q, idx_vecino[k][0], idx_vecino[k][1], idx_vecino[k][2],
                &utm[k][0], &utm[k][1], &utm[k][2]);
        k++ ;
    }

    // Compruebo vecino i-1, j, z
    if ((qi-1) >= 0)
    {
        idx_vecino[k][0] = qi-1;
        idx_vecino[k][1] = qj;
        idx_vecino[k][2] = qk;
        Utm3DHARM(q, idx_vecino[k][0], idx_vecino[k][1], idx_vecino[k][2],
                &utm[k][0], &utm[k][1], &utm[k][2]);
        k++ ;
    }

    // Compruebo vecino i, j+1, z
    if ((qj+1) < dimx)
    {
        idx_vecino[k][0] = qi;
        idx_vecino[k][1] = qj+1;
        idx_vecino[k][2] = qk;
        Utm3DHARM(q, idx_vecino[k][0], idx_vecino[k][1], idx_vecino[k][2],
                &utm[k][0], &utm[k][1], &utm[k][2]);
        k++;
    }

    // Compruebo vecino i, j-1, z
    if ((qj-1) >= 0)
    {
        idx_vecino[k][0] = qi;
        idx_vecino[k][1] = qj-1;
        idx_vecino[k][2] = qk;
        Utm3DHARM(q, idx_vecino[k][0], idx_vecino[k][1], idx_vecino[k][2],
                &utm[k][0], &utm[k][1], &utm[k][2]);
        k++;
    }

    // Compruebo vecino i, j, z+1
    if ((qk+1) < dimz)
    {
        idx_vecino[k][0] = qi;
        idx_vecino[k][1] = qj;
        idx_vecino[k][2] = qk+1;
        Utm3DHARM(q, idx_vecino[k][0], idx_vecino[k][1], idx_vecino[k][2],
                &utm[k][0], &utm[k][1], &utm[k][2]);
        k++;
    }

    // Compruebo vecino i, j, z-1
    if ((qk-1) >= 0)
    {
        idx_vecino[k][0] = qi;
        idx_vecino[k][1] = qj;
        idx_vecino[k][2] = qk-1;
        Utm3DHARM(q, idx_vecino[k][0], idx_vecino[k][1], idx_vecino[k][2],
                &utm[k][0], &utm[k][1], &utm[k][2]);
        k++;
    }


    DEBUG_INFO(
                const double inv_cte = 1.0/cte;
                fprintf(stdout, "InterpolaComponente7VecinosHARM:El pto mas alejado del pto dado contribuye con un factor %f\n",
                       SplineCubica(&inv_cte));
            fprintf(stdout, "InterpolaComponente7VecinosHARM: Encontrados %d puntos para interpolar\n", k);
    for (i=0; i<k; i++)
    {
        int qi1 = idx_vecino[i][0];
        int qj1 = idx_vecino[i][1];
        int qk1 = idx_vecino[i][2];
        int offset_xy = (qi1*q->dimx + qj1);
        int offset_xyz = (qk1*q->dimx*q->dimy) + offset_xy;
        double x1 = q->longitud[offset_xy];
        double y1 = q->latitud[offset_xy];
        double z1 = q->altura[offset_xyz];
        double u = q->data_u[offset_xyz];
        double v = q->data_v[offset_xyz];
        double w = q->data_w[offset_xyz];
        fprintf(stdout, "vecino %d: ", i);
        fprintf(stdout, "%d %d %d   utm=%f %f %f   u,v,w=%f %f %f\n",
               qi1, qj1, qk1, x1, y1, z1, u, v, w);
    }
    );

    // Me aseguro de que ha encontrado al menos tres vecinos
    assert(k > 2);

    // Buscar la distancia maxima del punto dado a sus vecinos -> dmax
    double dmax = -1.0;
    for (i=0; i < k; i++)
    {
        double dist = DistanciaR3(x, y, z, &utm[i][0], &utm[i][1], &utm[i][2]);
        dmax = (dist > dmax) ? dist : dmax ;
    }

    DEBUG_INFO(
                fprintf(stdout, "InterpolaComponente7VecinosHARM: dist. max. del pto a sus vecinos %f\n", dmax);
            );

    // Establecer el valor del radio r -> 2*dmax
    double r = cte * dmax;

    // Calcular el denominador
    double den = 0.0;
    for (i=0; i < k; i++)
    {
        double s = DistanciaR3(x, y, z, &utm[i][0], &utm[i][1], &utm[i][2]);
        double aux = s/r;
        den += SplineCubica(&aux);
    }

    // Calcular la interpolacion
    int p;
    double u, v, w;
    u = v = w = 0.0;
    for (p=0; p < k; p++)
    {
        double s = DistanciaR3(x, y, z, &utm[p][0], &utm[p][1], &utm[p][2]);
        double aux = s/r;
        double spline = SplineCubica(&aux);
        int qi1 = idx_vecino[p][0];
        int qj1 = idx_vecino[p][1];
        int qk1 = idx_vecino[p][2];
        int offset = (qk1 * q->dimx * q->dimy) +  qi1 * q->dimx + qj1;
        u = u + spline * q->data_u[offset];
        v = v + spline * q->data_v[offset];
        w = w + spline * q->data_w[offset];
    }
    wind[0] = u / den;
    wind[1] = v / den;
    wind[2] = w / den;
}

void Interpola7VecinosHARM(const struct VariableHARMONIE *q,
                           const double* puntox, const double* puntoy, const double* puntoz,
                           double *vx, double *vy, double *vz)
{
    double wind[3];
    InterpolaComponente7VecinosHARM(q, puntox, puntoy, puntoz, wind);
    *vx = wind[0];
    *vy = wind[1];
    *vz = wind[2];
}

