#include "localizador.h"

#include "avs.h"
#include "comparisons.h"
#include "debugInfo.h"
#include "triangleOverlap.h"

#include <stdlib.h>
#include <math.h>

/**
 * @file localizador.c
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

void getLocation_tetrahedron(const size_t nx, const size_t ny, const size_t nz,
                             const double x_min, const double y_min, const double z_min,
                             const double x_max, const double y_max, const double z_max,
                             double **coor, size_t **conec, const size_t numTetra,
                             struct sLocation3D *situation)
{
    size_t nx_sit = (size_t)floor(nx/5.0);
    size_t ny_sit = (size_t)floor(ny/5.0);
    size_t nz_sit = (size_t)floor(nz/5.0);

    situation->nx = nx_sit;
    situation->ny = ny_sit;
    situation->nz = nz_sit;
    situation->size = (size_t ***)malloc(nx_sit * sizeof(size_t**));
    situation->xMax = x_max;
    situation->yMax = y_max;
    situation->zMax = z_max;
    situation->xMin = x_min;
    situation->yMin = y_min;
    situation->zMin = z_min;

    for (size_t i = 0; i < nx_sit; i++) {
        situation->size[i] = (size_t **)malloc( ny_sit * sizeof(size_t *));
        for (size_t j=0; j<ny_sit; j++) {
            situation->size[i][j] = (size_t *)malloc( ny_sit * sizeof(size_t));
            for (size_t k=0; k<nz_sit; k++) {
                situation->size[i][j][k]=0;
            }
        }
    }


    for (size_t i=0; i<numTetra; i++) {

        double dx_sit = (x_max-x_min+1)/nx_sit;
        double dy_sit = (y_max-y_min+1)/ny_sit;
        double dz_sit = (z_max-z_min+1)/nz_sit;

        int nodo1 = conec[i][0];
        int nodo2 = conec[i][1];
        int nodo3 = conec[i][2];
        int nodo4 = conec[i][3];

        double x1 = coor[nodo1][0];
        double y1 = coor[nodo1][1];
        double z1 = coor[nodo1][2];

        double x2 = coor[nodo2][0];
        double y2 = coor[nodo2][1];
        double z2 = coor[nodo2][2];

        double x3 = coor[nodo3][0];
        double y3 = coor[nodo3][1];
        double z3 = coor[nodo3][2];

        double x4 = coor[nodo4][0];
        double y4 = coor[nodo4][1];
        double z4 = coor[nodo4][2];

        double x1_trans = x1 - x_min;
        double y1_trans = y1 - y_min;
        double z1_trans = z1 - z_min;

        int row = (int)floor(x1_trans/dx_sit);
        int col = (int)floor(y1_trans/dy_sit);
        int height = (int)floor(z1_trans/dz_sit);

        situation->size[row][col][height] = situation->size[row][col][height]+1;

        double x2_trans = x2 - x_min;
        double y2_trans = y2 - y_min;
        double z2_trans = z2 - z_min;

        row = (int)floor(x2_trans/dx_sit);
        col = (int)floor(y2_trans/dy_sit);
        height = (int)floor(z2_trans/dz_sit);

        situation->size[row][col][height] = situation->size[row][col][height]+1;

        double x3_trans = x3 - x_min;
        double y3_trans = y3 - y_min;
        double z3_trans = z3 - z_min;

        row = (int)floor(x3_trans/dx_sit);
        col = (int)floor(y3_trans/dy_sit);
        height = (int)floor(z3_trans/dz_sit);

        situation->size[row][col][height] = situation->size[row][col][height]+1;

        double x4_trans = x4 - x_min;
        double y4_trans = y4 - y_min;
        double z4_trans = z4 - z_min;

        row = (int)floor(x4_trans/dx_sit);
        col = (int)floor(y4_trans/dy_sit);
        height = (int)floor(z4_trans/dz_sit);

        situation->size[row][col][height] = situation->size[row][col][height]+1;

    }

    situation->location = (size_t ****)malloc( nx_sit * sizeof(size_t***));
    for (size_t i = 0; i < nx_sit; i++) {
        situation->location[i] = (size_t ***)malloc( ny_sit * sizeof(size_t**));
        for (size_t j=0; j<ny_sit; j++) {
            situation->location[i][j] = (size_t**)malloc(nz_sit * sizeof(size_t*));
            for (size_t k=0; k<nz_sit; k++) {
                situation->location[i][j][k] = (size_t*)malloc(situation->size[i][j][k] * sizeof(size_t));
                situation->size[i][j][k]=0;
            }
        }
    }

    for (size_t i=0; i<numTetra; i++) {

        double dx_sit = (x_max-x_min+1)/nx_sit;
        double dy_sit = (y_max-y_min+1)/ny_sit;
        double dz_sit = (z_max-z_min+1)/nz_sit;

        int nodo1 = conec[i][0];
        int nodo2 = conec[i][1];
        int nodo3 = conec[i][2];
        int nodo4 = conec[i][3];

        double x1 = coor[nodo1][0];
        double y1 = coor[nodo1][1];
        double z1 = coor[nodo1][2];

        double x2 = coor[nodo2][0];
        double y2 = coor[nodo2][1];
        double z2 = coor[nodo2][2];

        double x3 = coor[nodo3][0];
        double y3 = coor[nodo3][1];
        double z3 = coor[nodo3][2];

        double x4 = coor[nodo4][0];
        double y4 = coor[nodo4][1];
        double z4 = coor[nodo4][2];

        double x1_trans = x1 - x_min;
        double y1_trans = y1 - y_min;
        double z1_trans = z1 - z_min;

        int row = (int)floor(x1_trans/dx_sit);
        int col = (int)floor(y1_trans/dy_sit);
        int height = (int)floor(z1_trans/dz_sit);

        situation->location[row][col][height][situation->size[row][col][height]] = i;
        situation->size[row][col][height] = situation->size[row][col][height]+1;

        double x2_trans = x2 - x_min;
        double y2_trans = y2 - y_min;
        double z2_trans = z2 - z_min;

        row = (int)floor(x2_trans/dx_sit);
        col = (int)floor(y2_trans/dy_sit);
        height = (int)floor(z2_trans/dz_sit);

        situation->location[row][col][height][situation->size[row][col][height]] = i;
        situation->size[row][col][height] = situation->size[row][col][height]+1;

        double x3_trans = x3 - x_min;
        double y3_trans = y3 - y_min;
        double z3_trans = z3 - z_min;

        row = (int)floor(x3_trans/dx_sit);
        col = (int)floor(y3_trans/dy_sit);
        height = (int)floor(z3_trans/dz_sit);

        situation->location[row][col][height][situation->size[row][col][height]] = i;
        situation->size[row][col][height] = situation->size[row][col][height]+1;

        double x4_trans = x4 - x_min;
        double y4_trans = y4 - y_min;
        double z4_trans = z4 - z_min;

        row = (int)floor(x4_trans/dx_sit);
        col = (int)floor(y4_trans/dy_sit);
        height = (int)floor(z4_trans/dz_sit);

        situation->location[row][col][height][situation->size[row][col][height]] = i;
        situation->size[row][col][height] = situation->size[row][col][height]+1;

    }
}


void getLocation_triangle(const size_t nx, const size_t ny, const double x_min, const double y_min,
                          const double x_max, const double y_max, double **coor, size_t **conec,
                          const size_t numCaras, struct sLocation2D *situation)
{

    size_t nx_sit = (size_t)floor(nx/5.0);
    size_t ny_sit = (size_t)floor(ny/5.0);

    situation->nx = nx_sit;
    situation->ny = ny_sit;
    situation->size = (size_t **)malloc(nx_sit * sizeof(size_t*));
    situation->xMax = x_max;
    situation->yMax = y_max;
    situation->xMin = x_min;
    situation->yMin = y_min;

    for (size_t i = 0; i < nx_sit; i++) {
        situation->size[i] = (size_t *)malloc( ny_sit * sizeof(size_t));
        for (size_t j=0; j<ny_sit; j++) {
            situation->size[i][j]=0;
        }
    }


    for (size_t i=0; i<numCaras; i++) {

        double dx_sit = (x_max-x_min+1)/nx_sit;
        double dy_sit = (y_max-y_min+1)/ny_sit;

        int nodo1 = conec[i][0];
        int nodo2 = conec[i][1];
        int nodo3 = conec[i][2];

        double x1 = coor[nodo1][0];
        double y1 = coor[nodo1][1];

        double x2 = coor[nodo2][0];
        double y2 = coor[nodo2][1];

        double x3 = coor[nodo3][0];
        double y3 = coor[nodo3][1];

        double x1_trans = x1 - x_min;
        double y1_trans = y1 - y_min;

        int row = (int)floor(x1_trans/dx_sit);
        int col = (int)floor(y1_trans/dy_sit);

        situation->size[row][col] = situation->size[row][col]+1;

        double x2_trans = x2 - x_min;
        double y2_trans = y2 - y_min;

        row = (int)floor(x2_trans/dx_sit);
        col = (int)floor(y2_trans/dy_sit);

        situation->size[row][col] = situation->size[row][col]+1;

        double x3_trans = x3 - x_min;
        double y3_trans = y3 - y_min;

        row = (int)floor(x3_trans/dx_sit);
        col = (int)floor(y3_trans/dy_sit);

        situation->size[row][col] = situation->size[row][col]+1;

    }

    situation->location = (size_t ***)malloc( nx_sit * sizeof(size_t**));
    for (size_t i = 0; i < nx_sit; i++) {
        situation->location[i] = (size_t **)malloc( ny_sit * sizeof(size_t*));
        for (size_t j=0; j<ny_sit; j++) {
            situation->location[i][j] = (size_t*)malloc(situation->size[i][j] * sizeof(size_t));
            situation->size[i][j]=0;
        }
    }

    for (size_t i=0; i<numCaras; i++) {

        double dx_sit = (x_max-x_min+1)/nx_sit;
        double dy_sit = (y_max-y_min+1)/ny_sit;

        int nodo1 = conec[i][0];
        int nodo2 = conec[i][1];
        int nodo3 = conec[i][2];

        double x1 = coor[nodo1][0];
        double y1 = coor[nodo1][1];

        double x2 = coor[nodo2][0];
        double y2 = coor[nodo2][1];

        double x3 = coor[nodo3][0];
        double y3 = coor[nodo3][1];

        double x1_trans = x1 - x_min;
        double y1_trans = y1 - y_min;

        int row = (int)floor(x1_trans/dx_sit);
        int col = (int)floor(y1_trans/dy_sit);

        situation->location[row][col][situation->size[row][col]] = i;
        situation->size[row][col] = situation->size[row][col]+1;

        double x2_trans = x2 - x_min;
        double y2_trans = y2 - y_min;

        row = (int)floor(x2_trans/dx_sit);
        col = (int)floor(y2_trans/dy_sit);

        situation->location[row][col][situation->size[row][col]] = i;
        situation->size[row][col] = situation->size[row][col]+1;

        double x3_trans = x3 - x_min;
        double y3_trans = y3 - y_min;

        row = (int)floor(x3_trans/dx_sit);
        col = (int)floor(y3_trans/dy_sit);

        situation->location[row][col][situation->size[row][col]] = i;
        situation->size[row][col] = situation->size[row][col]+1;


    }
}

size_t getTetrahedra(const double x, const double y, const double z,
                     const double **coord, const size_t **conec, const size_t numberElems,
                     const struct sLocation3D *situation)
{
    size_t nx_sit = situation->nx;
    size_t ny_sit = situation->ny;
    size_t nz_sit = situation->nz;

    double dx_sit = (situation->xMax-situation->xMin+1)/nx_sit;
    double dy_sit = (situation->yMax-situation->yMin+1)/ny_sit;
    double dz_sit = (situation->zMax-situation->zMin+1)/nz_sit;

    double punto[3] = {x,y,z};

    double x_trans = x - situation->xMin;
    double y_trans = y - situation->yMin;
    double z_trans = z - situation->zMin;

    size_t row = (int)floor(x_trans/dx_sit);
    size_t col = (int)floor(y_trans/dy_sit);
    size_t height = (int)floor(z_trans/dz_sit);

    for (size_t j = 0; j < situation->size[row][col][height]; j++) {
        size_t tetraedro = situation->location[row][col][height][j];

        size_t idx_nodo1 = conec[tetraedro][0];
        size_t idx_nodo2 = conec[tetraedro][1];
        size_t idx_nodo3 = conec[tetraedro][2];
        size_t idx_nodo4 = conec[tetraedro][3];

        double nodo1[3] = {coord[idx_nodo1][0],
                           coord[idx_nodo1][1],
                           coord[idx_nodo1][2]};
        double nodo2[3] = {coord[idx_nodo2][0],
                           coord[idx_nodo2][1],
                           coord[idx_nodo2][2]};
        double nodo3[3] = {coord[idx_nodo3][0],
                           coord[idx_nodo3][1],
                           coord[idx_nodo3][2]};
        double nodo4[3] = {coord[idx_nodo4][0],
                           coord[idx_nodo4][1],
                           coord[idx_nodo4][2]};

        if (point_inside_tetrahedra(nodo1, nodo2, nodo3, nodo4, punto)) {
            return tetraedro;
        }
    }


    for (int row_inc = -1; row_inc < 2; row_inc++) {
        if ((row == 0) && (row_inc == -1)) {
            continue;
        }
        size_t row_vecinos = row + row_inc;
        if (row_vecinos >= nx_sit) {
            continue;
        }
        for (int col_inc = -1; col_inc < 2; col_inc++) {
            if ((col == 0) && (col_inc == -1)) {
                continue;
            }
            size_t col_vecinos = col + col_inc;
            if (col_vecinos >= ny_sit) {
                continue;
            }
            for (int height_inc = -1; height_inc < 2; height_inc++) {

                if ((height == 0) && (height_inc == -1))
                {
                    continue;
                }
                size_t height_vecinos = height + height_inc;
                if (height_vecinos > nz_sit)
                {
                    continue;
                }
                if ( (col_vecinos == col) && (row_vecinos == row) && (height_vecinos == height)) {
                    continue;
                }

                for (size_t j = 0; j < situation->size[row_vecinos][col_vecinos][height_vecinos]; j++) {
                    size_t tetraedro = situation->location[row_vecinos][col_vecinos][height_vecinos][j];

                    size_t idx_nodo1 = conec[tetraedro][0];
                    size_t idx_nodo2 = conec[tetraedro][1];
                    size_t idx_nodo3 = conec[tetraedro][2];
                    size_t idx_nodo4 = conec[tetraedro][3];

                    double nodo1[3] = {coord[idx_nodo1][0],
                                       coord[idx_nodo1][1],
                                       coord[idx_nodo1][2]};
                    double nodo2[3] = {coord[idx_nodo2][0],
                                       coord[idx_nodo2][1],
                                       coord[idx_nodo2][2]};
                    double nodo3[3] = {coord[idx_nodo3][0],
                                       coord[idx_nodo3][1],
                                       coord[idx_nodo3][2]};
                    double nodo4[3] = {coord[idx_nodo4][0],
                                       coord[idx_nodo4][1],
                                       coord[idx_nodo4][2]};

                    if (point_inside_tetrahedra(nodo1, nodo2, nodo3, nodo4, punto)) {
                        return tetraedro;
                    }
                }
            }
        }
    }



    fprintf(stderr, "No se ha encontrado tetraedro");
    fprintf(stderr, "Punto: %f %f %f\n", x, y, z);
    fprintf(stderr, "row: %zu, colum: %zu\n, layer: %zu\n", row, col, height);

    return(numberElems+1);
}

size_t getTriangle(const double x, const double y,
                   const double **coord, const size_t **conec,
                   const size_t numElems,
                   const struct sLocation2D *situation)
{

    size_t nx_sit = situation->nx;
    size_t ny_sit = situation->ny;

    double dx_sit = (situation->xMax-situation->xMin+1)/nx_sit;
    double dy_sit = (situation->yMax-situation->yMin+1)/ny_sit;

    double punto[2] = {x,y};

    double x_trans = x - situation->xMin;
    double y_trans = y - situation->yMin;

    size_t row = (int)floor(x_trans/dx_sit);
    size_t col = (int)floor(y_trans/dy_sit);

    for (size_t j = 0; j < situation->size[row][col]; j++) {
        size_t cara = situation->location[row][col][j];

        size_t idx_nodo1 = conec[cara][0];
        size_t idx_nodo2 = conec[cara][1];
        size_t idx_nodo3 = conec[cara][2];

        double nodo1[3] = {coord[idx_nodo1][0],
                           coord[idx_nodo1][1],
                           coord[idx_nodo1][2]};
        double nodo2[3] = {coord[idx_nodo2][0],
                           coord[idx_nodo2][1],
                           coord[idx_nodo2][2]};
        double nodo3[3] = {coord[idx_nodo3][0],
                           coord[idx_nodo3][1],
                           coord[idx_nodo3][2]};

        if (point_inside_triangle(nodo1, nodo2, nodo3, punto)) {
            return cara;
        }
    }


    for (int row_inc = -1; row_inc < 2; row_inc++) {
        if ((row == 0) && (row_inc == -1)) {
            continue;
        }
        size_t row_vecinos = row + row_inc;
        if (row_vecinos >= nx_sit) {
            continue;
        }
        for (int col_inc = -1; col_inc < 2; col_inc++) {
            if ((col == 0) && (col_inc == -1)) {
                continue;
            }
            size_t col_vecinos = col + col_inc;
            if (col_vecinos >= ny_sit) {
                continue;
            }
            if ( (col_vecinos == col) && (row_vecinos == row)) {
                continue;
            }

            for (size_t j = 0; j < situation->size[row_vecinos][col_vecinos]; j++) {
                size_t cara = situation->location[row_vecinos][col_vecinos][j];

                size_t idx_nodo1 = conec[cara][0];
                size_t idx_nodo2 = conec[cara][1];
                size_t idx_nodo3 = conec[cara][2];

                double nodo1[3] = {coord[idx_nodo1][0],
                                   coord[idx_nodo1][1],
                                   coord[idx_nodo1][2]};
                double nodo2[3] = {coord[idx_nodo2][0],
                                   coord[idx_nodo2][1],
                                   coord[idx_nodo2][2]};
                double nodo3[3] = {coord[idx_nodo3][0],
                                   coord[idx_nodo3][1],
                                   coord[idx_nodo3][2]};


                if (point_inside_triangle(nodo1, nodo2, nodo3, punto)) {
                    return cara;
                }
            }
        }
    }



    fprintf(stderr, "No se ha encontrado triangulo\n");
    fprintf(stderr, "Punto: %f %f\n", x, y);
    fprintf(stderr, "row: %zu, colum: %zu\n", row, col);

    return(numElems+1);

}

bool point_inside_tetrahedra(const double *p1, const double *p2,
                             const double *p3, const double *p4,
                             const double *punto)
{
    const double vgrande = VolTetra (&p1[0], &p1[1], &p1[2],
            &p2[0], &p2[1], &p2[2],
            &p3[0], &p3[1], &p3[2],
            &p4[0], &p4[1], &p4[2]);

    if (equalToZero(&vgrande)){
        fprintf(stderr, "Tetraedro degenerado!!!\n");
    }

    // Calcula los volumenes de cada tetraedro formado por el punto (X,Y,Z) y tres
    // de los nodos del tetraedro. Si el volumen de un tetraedro es igual a 0
    // significa que el punto está en una arista. Si el signo del volumen es contrario
    // al del volumen total, el punto está fuera
    const double v1 = VolTetra (&punto[0], &punto[1], &punto[2],
            &p2[0], &p2[1], &p2[2],
            &p3[0], &p3[1], &p3[2],
            &p4[0], &p4[1], &p4[2]);

    //Point is in the boundary
    if (equalToZero(&v1)){
        return true;
    }
    //Check sign, if it differs it is outside
    double mult = v1 * vgrande;
    if (lessThanZero(&mult)){
        return false;
    }

    const double v2 = VolTetra (&p1[0], &p1[1], &p1[2],
            &punto[0], &punto[1], &punto[2],
            &p3[0], &p3[1], &p3[2],
            &p4[0], &p4[1], &p4[2]);
    //Point is in the boundary
    if (equalToZero(&v2)){
        return true;
    }
    //Check sign, if it differs it is outside
    mult = v2 * vgrande;
    if (lessThanZero(&mult)){
        return false;
    }

    const double v3 = VolTetra(&p1[0], &p1[1], &p1[2],
            &p2[0], &p2[1], &p2[2],
            &punto[0], &punto[1], &punto[2],
            &p4[0], &p4[1], &p4[2]);
    //Point is in the boundary
    if (equalToZero(&v3)){
        return true;
    }
    //Check sign, if it differs it is outside
    mult = v3 * vgrande;
    if (lessThanZero(&mult)){
        return false;
    }

    const double v4 = VolTetra (&p1[0], &p1[1], &p1[2],
            &p2[0], &p2[1], &p2[2],
            &p3[0], &p3[1], &p3[2],
            &punto[0], &punto[1], &punto[2]);
    //Point is in the boundary
    if (equalToZero(&v4)){
        return true;
    }
    //Check sign, if it differs it is outside
    mult = v4 * vgrande;
    if (lessThanZero(&mult)){
        return false;
    }

    // si llega aquí es que el punto es interior al tetraedro.

    DEBUG_INFO(
                //Comprobamos que la suma de los 4 subvolumenes es igual al volumen total
                const double aux = v1 + v2 + v3 + v4;
            if (!equal(&aux,&vgrande)){
        fprintf(stderr,"Pasa algo raro en point_inside_tetrahedra\n");
        fprintf(stderr, "Valor de vgrande: %g\n", vgrande);
        fprintf(stderr, "Valor de vaux:    %g\n", aux);
        fprintf(stderr, "Diferencia:       %g\n", vgrande-aux);
    }
    );


    return true;

}

bool point_inside_triangle(const double *p1, const double *p2, const double *p3, const double *punto)
{
    const double V1[2] = {p2[0]-p1[0], p2[1]-p1[1]};
    const double V2[2] = {p3[0]-p1[0], p3[1]-p1[1]};
    const double V3[2] = {punto[0]-p1[0], punto[1]-p1[1]};

    const double dot11 = (V1[0] * V1[0]) + (V1[1] * V1[1]);
    const double dot12 = (V1[0] * V2[0]) + (V1[1] * V2[1]);
    const double dot13 = (V1[0] * V3[0]) + (V1[1] * V3[1]);
    const double dot22 = (V2[0] * V2[0]) + (V2[1] * V2[1]);
    const double dot23 = (V2[0] * V3[0]) + (V2[1] * V3[1]);

    double inv_denom = (dot11*dot22) - (dot12*dot12);

    if (equalToZero(&inv_denom)) {
        return(false);
    }

    inv_denom = 1.0/inv_denom;

    const double u = ((dot22 * dot13) - (dot12 * dot23)) * inv_denom;
    const double v = ((dot11 * dot23) - (dot12 * dot13)) * inv_denom;
    const double w = 1-(u + v);

    return( greatThanOrEqualToZero(&u) &&
            greatThanOrEqualToZero(&v) &&
            greatThanOrEqualToZero(&w));
}


void set_bb_branches(struct tree *situation)
{
    const double x_centre = situation->x_min + ((situation->x_max-situation->x_min)/2.0);
    const double y_centre = situation->y_min + ((situation->y_max-situation->y_min)/2.0);

    situation->branch[0].x_min = situation->x_min;
    situation->branch[0].y_min = situation->y_min;
    situation->branch[0].x_max = x_centre;
    situation->branch[0].y_max = y_centre;


    situation->branch[1].x_min = x_centre;
    situation->branch[1].y_min = situation->y_min;
    situation->branch[1].x_max = situation->x_max;
    situation->branch[1].y_max = y_centre;

    situation->branch[2].x_min = situation->x_min;
    situation->branch[2].y_min = y_centre;
    situation->branch[2].x_max = x_centre;
    situation->branch[2].y_max = situation->y_max;

    situation->branch[3].x_min = x_centre;
    situation->branch[3].y_min = y_centre;
    situation->branch[3].x_max = situation->x_max;
    situation->branch[3].y_max = situation->y_max;

    for(size_t i=0; i<4; ++i){
        situation->branch[i].z_max = situation->z_max;
        situation->branch[i].z_min = situation->z_min;
    }


}

void set_bb_branches_OT(struct tree *situation)
{
    const double x_centre = situation->x_min + ((situation->x_max-situation->x_min)/2.0);
    const double y_centre = situation->y_min + ((situation->y_max-situation->y_min)/2.0);
    const double z_centre = situation->z_min + ((situation->z_max-situation->z_min)/2.0);

    situation->branch[0].x_min = situation->x_min;
    situation->branch[0].y_min = situation->y_min;
    situation->branch[0].z_min = situation->z_min;
    situation->branch[0].x_max = x_centre;
    situation->branch[0].y_max = y_centre;
    situation->branch[0].z_max = z_centre;

    situation->branch[1].x_min = x_centre;
    situation->branch[1].y_min = situation->y_min;
    situation->branch[1].z_min = situation->z_min;
    situation->branch[1].x_max = situation->x_max;
    situation->branch[1].y_max = y_centre;
    situation->branch[1].z_max = z_centre;

    situation->branch[2].x_min = situation->x_min;
    situation->branch[2].y_min = y_centre;
    situation->branch[2].z_min = situation->z_min;
    situation->branch[2].x_max = x_centre;
    situation->branch[2].y_max = situation->y_max;
    situation->branch[2].z_max = z_centre;

    situation->branch[3].x_min = x_centre;
    situation->branch[3].y_min = y_centre;
    situation->branch[3].z_min = situation->z_min;
    situation->branch[3].x_max = situation->x_max;
    situation->branch[3].y_max = situation->y_max;
    situation->branch[3].z_max = z_centre;

    situation->branch[4].x_min = situation->x_min;
    situation->branch[4].y_min = situation->y_min;
    situation->branch[4].z_min = z_centre;
    situation->branch[4].x_max = x_centre;
    situation->branch[4].y_max = y_centre;
    situation->branch[4].z_max = situation->z_max;

    situation->branch[5].x_min = x_centre;
    situation->branch[5].y_min = situation->y_min;
    situation->branch[5].z_min = z_centre;
    situation->branch[5].x_max = situation->x_max;
    situation->branch[5].y_max = y_centre;
    situation->branch[5].z_max = situation->z_max;

    situation->branch[6].x_min = situation->x_min;
    situation->branch[6].y_min = y_centre;
    situation->branch[6].z_min = z_centre;
    situation->branch[6].x_max = x_centre;
    situation->branch[6].y_max = situation->y_max;
    situation->branch[6].z_max = situation->z_max;

    situation->branch[7].x_min = x_centre;
    situation->branch[7].y_min = y_centre;
    situation->branch[7].z_min = z_centre;
    situation->branch[7].x_max = situation->x_max;
    situation->branch[7].y_max = situation->y_max;
    situation->branch[7].z_max = situation->z_max;

}

bool tringleIsInsideQT(struct tree *situation, const double **coor, const size_t *conec)
{

    double boxHalfSize[3] = {(situation->x_max-situation->x_min)/2.0,
                             (situation->y_max-situation->y_min)/2.0,
                             (situation->z_max-situation->z_min)/2.0,};

    double boxCentre[3]  = {situation->x_min + boxHalfSize[0],
                            situation->y_min + boxHalfSize[1],
                            situation->z_min + boxHalfSize[2]};

    double verts[3][3];

    //    fprintf(stdout, "element: %zu\n", element);
    //    fflush(stdout);
    for (size_t i=0; i < 3; ++i){
        size_t nodo = conec[i];
        for (size_t j=0; j < 3; ++j){
            verts[i][j] = coor[nodo][j];
        }
    }

    return triBoxOverlap(boxCentre, boxHalfSize, verts);
}

void passElements(const size_t **conec, const double **coor,
                  struct tree *situation, const size_t i)
{

    size_t *temporary_elems = (size_t*)malloc(situation->nOfElements*sizeof(size_t));
    size_t temporary_nOfElements = 0;
    if (temporary_elems==NULL){
        fprintf(stderr, "Ha habido un fallo alocatando temporary_elements\n");
        exit(EXIT_FAILURE);
    }

    for (size_t j=0; j<situation->nOfElements; ++j){
        if (tringleIsInsideQT(&situation->branch[i], coor, conec[situation->elements[j]])){
            temporary_elems[temporary_nOfElements] = situation->elements[j];
            ++temporary_nOfElements;
        }
    }

    situation->branch[i].elements = (size_t *)malloc(temporary_nOfElements*sizeof(size_t));
    if (situation->branch[i].elements == NULL){
        fprintf(stderr, "No se ha podido alocatar situation->branch[%zu].elements\n", i);
        exit(EXIT_FAILURE);
    }

    situation->branch[i].nOfElements = temporary_nOfElements;
    for (size_t j=0; j<temporary_nOfElements; ++j){
        situation->branch[i].elements[j] = temporary_elems[j];
    }

    free(temporary_elems);
}

void inheritParametres(struct tree *situation, const size_t i)
{
    situation->branch[i].maxNOfElements = situation->maxNOfElements;
    situation->branch[i].maxLevel = situation->maxLevel;
    situation->branch[i].isRoot = false;
    situation->branch[i].branch = NULL;
    situation->branch[i].nOfElements = 0;
    situation->branch[i].level = situation->level+1;
}

void refine_quadTree(struct tree *situation, const double **coor, const size_t **conec)
{
    if (situation->nOfElements <= situation->maxNOfElements
            ||  situation->level == situation->maxLevel) {
        return;
    }

    situation->isRoot = true;

    situation->branch = (struct tree*)malloc(4*sizeof(struct tree));

    if (situation->branch == NULL){
        fprintf(stderr, "No se ha podido alocatar situation->branch\n");
        exit(EXIT_FAILURE);
    }

    for (size_t i=0; i<4; ++i){
        inheritParametres(situation, i);
    }

    set_bb_branches(situation);

    for (size_t i=0; i<4; ++i){
        passElements(conec, coor, situation, i);
        refine_quadTree(&(situation->branch[i]), coor, conec);
    }

    free(situation->elements);
    situation->nOfElements = 0;


}

void getLocation_quadTree(const double x_min, const double y_min, const double z_min,
                          const double x_max, const double y_max, const double z_max,
                          const double **coor, const size_t **conec,
                          const size_t numCaras, struct tree *situation)
{


    situation->x_max = x_max;
    situation->x_min = x_min;
    situation->y_max = y_max;
    situation->y_min = y_min;
    situation->z_max = z_max;
    situation->z_min = z_min;

    situation->maxNOfElements = 8;
    situation->maxLevel = 32;
    situation->isRoot = false;
    situation->branch = NULL;
    situation->nOfElements = 0;
    situation->level = 0;

    situation->elements = (size_t*)malloc(numCaras*sizeof(size_t));
    if (situation->elements == NULL) {
        fprintf(stderr, "No se ha podido alocatar bien situation->elements\n");
        exit(EXIT_FAILURE);
    }

    situation->nOfElements = numCaras;

    for (size_t i=0; i < numCaras; ++i)
    {
        situation->elements[i] = i;
    }

    refine_quadTree(situation, coor, conec);

}

void getElementsQT(const double x, const double y, const struct tree *situation, size_t **elements, size_t *nOfElements)
{
    if (situation->isRoot){
        size_t ind1=0;
        double x_centre = situation->x_min + ((situation->x_max-situation->x_min)/2.0);
        double y_centre = situation->y_min + ((situation->y_max-situation->y_min)/2.0);
        if (greatThan(&x,&x_centre)) {
            ind1 = 1;
        }
        if (greatThan(&y,&y_centre)) {
            ind1 = ind1+2;
        }
        getElementsQT(x, y, &situation->branch[ind1], elements, nOfElements);
    } else {

        *nOfElements = situation->nOfElements;
        *elements = situation->elements;
    }

}

size_t getTriangle_QT(const double x, const double y,
                      const double **coord, const size_t **conec, const size_t numElems,
                      const struct tree *situation)
{
    size_t * elements  = NULL;
    size_t nOfElements;
    getElementsQT(x, y, situation, &elements, &nOfElements);

    double punto[2] = {x,y};
    for(size_t nOfCara=0; nOfCara < nOfElements; ++nOfCara){

        size_t cara = elements[nOfCara];
        size_t idx_nodo1 = conec[cara][0];
        size_t idx_nodo2 = conec[cara][1];
        size_t idx_nodo3 = conec[cara][2];

        double nodo1[3] = {coord[idx_nodo1][0],
                           coord[idx_nodo1][1],
                           coord[idx_nodo1][2]};
        double nodo2[3] = {coord[idx_nodo2][0],
                           coord[idx_nodo2][1],
                           coord[idx_nodo2][2]};
        double nodo3[3] = {coord[idx_nodo3][0],
                           coord[idx_nodo3][1],
                           coord[idx_nodo3][2]};

        if (point_inside_triangle(nodo1, nodo2, nodo3, punto)) {
            return cara;
        }
    }


    return(numElems+1);
}

void freeQT(struct tree *situation)
{
    if (situation->isRoot){
        for (size_t i = 0; i<4; ++i){
            freeQT(&(situation->branch[i]));
        }
    } else {
        free(situation->elements);
    }

    free(situation->branch);
}

void passElementsOT(const size_t **conec, const double **coor, struct tree *situation, size_t i)
{

    size_t *temporary_elems = (size_t*)malloc(situation->nOfElements*sizeof(size_t));
    size_t temporary_nOfElements = 0;
    if (temporary_elems==NULL){
        fprintf(stderr, "Ha habido un fallo alocatando temporary_elements\n");
        exit(EXIT_FAILURE);
    }

    for (size_t j=0; j<situation->nOfElements; ++j){
        const size_t element = situation->elements[j];
        const size_t conec_local0[3] = {conec[element][0], conec[element][1], conec[element][2]};
        const size_t conec_local1[3] = {conec[element][0], conec[element][1], conec[element][2]};
        const size_t conec_local2[3] = {conec[element][0], conec[element][1], conec[element][2]};
        const size_t conec_local3[3] = {conec[element][0], conec[element][1], conec[element][2]};

        if (tringleIsInsideQT(&situation->branch[i], coor, conec_local0) ||
                tringleIsInsideQT(&situation->branch[i], coor, conec_local1) ||
                tringleIsInsideQT(&situation->branch[i], coor, conec_local2) ||
                tringleIsInsideQT(&situation->branch[i], coor, conec_local3)){
            temporary_elems[temporary_nOfElements] = situation->elements[j];
            ++temporary_nOfElements;
        }
    }

    situation->branch[i].elements = (size_t *)malloc(temporary_nOfElements*sizeof(size_t));
    if (situation->branch[i].elements == NULL){
        fprintf(stderr, "No se ha podido alocatar situation->branch[%zu].elements\n", i);
        exit(EXIT_FAILURE);
    }

    situation->branch[i].nOfElements = temporary_nOfElements;
    for (size_t j=0; j<temporary_nOfElements; ++j){
        situation->branch[i].elements[j] = temporary_elems[j];
    }

    free(temporary_elems);
}


void refine_OctTree(struct tree *situation, const double **coor, const size_t **conec)
{
    if (situation->nOfElements <= situation->maxNOfElements
            || situation->level == situation->maxLevel
            ) {
        return;
    }

    situation->isRoot = true;

    situation->branch = (struct tree*)malloc(8*sizeof(struct tree));

    if (situation->branch == NULL){
        fprintf(stderr, "No se ha podido alocatar situation->branch\n");
        exit(EXIT_FAILURE);
    }

    for (size_t i=0; i<8; ++i){
        inheritParametres(situation, i);
    }

    set_bb_branches_OT(situation);

    for (size_t i=0; i<8; ++i){
        passElementsOT(conec, coor, situation, i);
        refine_OctTree(&(situation->branch[i]), coor, conec);
    }

    free(situation->elements);
    situation->nOfElements = 0;


}

void getLocation_OctTree(const double x_min, const double y_min, const double z_min,
                         const double x_max, const double y_max, const double z_max,
                         const double **coor, const size_t **conec,
                         const size_t numElems, struct tree *situation)
{


    situation->x_max = x_max;
    situation->x_min = x_min;
    situation->y_max = y_max;
    situation->y_min = y_min;
    situation->z_max = z_max;
    situation->z_min = z_min;

    situation->maxNOfElements = 64;
    situation->maxLevel = 16;
    situation->isRoot = false;
    situation->branch = NULL;
    situation->nOfElements = 0;
    situation->level = 0;

    situation->elements = (size_t*)malloc(numElems*sizeof(size_t));
    if (situation->elements == NULL) {
        fprintf(stderr, "No se ha podido alocatar bien situation->elements\n");
        exit(EXIT_FAILURE);
    }

    situation->nOfElements = numElems;

    for (size_t i=0; i < numElems; ++i)
    {
        situation->elements[i] = i;
    }

    refine_OctTree(situation, coor, conec);

}
