/*********************************************************/
/* AABB-triangle overlap test code                       */
/* by Tomas Akenine-Möller                               */
/* Function: int triBoxOverlap(double boxcenter[3],      */
/*          double boxhalfsize[3],double triverts[3][3]) */
/* History:                                              */
/*   2001-03-05: released the code in its first version  */
/*   2001-06-18: changed the order of the tests, faster  */
/*                                                       */
/* Acknowledgement: Many thanks to Pierre Terdiman for   */
/* suggestions and discussions on how to optimize code.  */
/* Thanks to David Hunt for finding a ">="-bug!          */
/*********************************************************/

/**
 * @file triangleOverlap.h
 * @brief Fichero que usa el AABB-triangle overlap test code by Tomas
 * Akenine-Möller, para comprovar la intersección de triángulos y bounding box
 */

#include <stdbool.h>

/**
 * @brief planeBoxOverlap
 * @param[in] normal
 * @param[in] d
 * @param[in] maxbox
 * @return
 */
bool planeBoxOverlap(double normal[3], double d, double maxbox[3]);

/**
 * @brief Función que comprueba si un triangulo intersecta a un bounding box
 * @param[in] boxcenter Centro de la bounding box
 * @param[in] boxhalfsize Longitud media de la bounding box
 * @param[in] triverts Vector con las coordenadas de los tres vertices del triángulo
 * @return Verdadero si el triangulo intersecta al bounding box, falso en caso contrario
 */
bool triBoxOverlap(double boxcenter[3],double boxhalfsize[3],double triverts[3][3]);

