#ifndef WIND3D_H
#define WIND3D_H

#include "localizador.h"

#include <stdbool.h>
#include <stddef.h>

/**
 * @file wind3D.h
 * @brief Estructuras para el programa completo
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @brief Valor de pi
 */
#define PI 3.14159265358979323846

/**
 * @brief Struct que usaremos para definir la superfície del terreno
 */
struct sTerreno {
	/**
	 * @brief Num_caras Número de caras
	 */
	size_t Num_caras;
	/**
	 * @brief x_max Valor de la coordenada x máxima
	 */
	double x_max;
	/**
	 * @brief y_max Valor de la coordenada y máxima
	 */
	double y_max;
	/**
	 * @brief z_max Valor de la coordenada z máxima
	 */
	double z_max;
	/**
	 * @brief x_min Valor de la coordenada x mínima
	 */
	double x_min;
	/**
	 * @brief y_min Valor de la coordenada y mínima
	 */
	double y_min;
	/**
	 * @brief z_min Valor de la coordenada z mínima
	 */
	double z_min;
	/**
	 * @brief cara Puntero que apunta a las conectividades de la cara
	 */
	size_t **cara;
	/**
	 * @brief situation Puntero a la estructura de localización
	 *        rápida de los elementos del terreno
	 */
	//    struct sLocation2D situation;
	struct tree situation;
} ;

/**
 * @brief Estructura para guadar los datos de la malla
 */

struct T_Malla {
	/**
	 * @brief Número de nodos de la malla
	 */
	size_t Num_nodos;
	/**
	 * @brief Número de tetraedros de la malla
	 */
	size_t Num_tetra;
	/**
	 * @brief Puntero a vector de coordenadas de los nodos
	 */
	double **nodo;
	/**
	 * @brief Puntero a vector de alturas del terreno en la proyección del centro del tetraedro
	 */
	double *z_terr;
	/**
	 * @brief Puntero a número de condición de contorno de cada nodo
	 */
	short int *condicion;
	/**
	 * @brief vini Puntero al viento interpolado de cada elemento
	 */
	double **vini;
	/**
	 * @brief wind Puntero al viento resultante de aplicar masa consistente de cada elemento
	 */
	double **wind;
	/**
	 * @brief tetra Puntero a las conectividades de cada tetraedro
	 */
	size_t **tetra;
	/**
	 * @brief cen_tetra Puntero a las coordenadas del centro geométrico de cada tetraedro
	 */
	double **cen_tetra;
	/**
	 * @brief terreno Estructura donde se guarda la información del terreno de la malla
	 */
	struct sTerreno terreno;
	/**
	 * @brief situation Puntero a la estructura de localización
	 *        rápida de los elementos de la malla
	 */
	struct sLocation3D situation;
	/**
	 * @brief Puntero al nombre del fichero con los puntos de salida
	 */
	char *fichPuntosSalida;
	/**
	 * @brief Puntero al nombre del fichero con los puntos de control
	 */
	char *fichPuntosControl;
	/**
	 * @brief Booleano para conocer si se usan los puntos de salida
	 */
	bool puntosSalida;
	/**
	 * @brief Booleano para conocer si se usan los puntos de control
	 */
	bool puntosControl;
	/**
	 * @brief Booleano para conocer si se calculan distintas capas
	 */
	bool tieneCapas;
};

/**
 * @brief Estructura para representar la topografia
 */
struct T_Topografia{
	/**
	 * @brief Valor mínimo de la coordenada X
	 */
	double xmin;
	/**
	 * @brief Valor máximo de la coordenada X
	 */
	double xmax;
	/**
	 * @brief Valor mínimo de la coordenada Y
	 */
	double ymin;
	/**
	 * @brief Valor máximo de la coordenada Y
	 */
	double ymax;
	/**
	 * @brief Valor del tamaño de cuadrícula en la coordenada X
	 */
	double deltax;
	/**
	 * @brief Valor del tamaño de cuadrícula en la coordenada Y
	 */
	double deltay;
	/**
	 * @brief Número de intervalos en la coordenada X
	 */
	size_t nx;
	/**
	 * @brief Número de intervalos en la coordenada Y
	 */
	size_t ny;
	/**
	 * @brief Valores de la altura
	 */
	double **z;
	/**
	 * @brief Valores de la rugosidad
	 */
	double **z0;
};

/**
 * @brief Estructura para representar una estacion
 */
struct T_Estacion{
	/**
	 * @brief Coordenada X de la estación
	 */
	double x;
	/**
	 * @brief Coordenada Y de la estación
	 */
	double y;
	/**
	 * @brief Coordenada Z de la estación
	 */
	double z;
	/**
	 * @brief Altura sobre el terreno
	 */
	double z_terr;
	/**
	 * @brief Módulo de la velocidad del viento, en m/s
	 */
	double modulo;
	/**
	 * @brief   Ángulo del viento, en grados
	 */
	double angulo;
	/**
	 * @brief Componente X del viento
	 */
	double vx;
	/**
	 * @brief Componente Y del viento
	 */
	double vy;
	/**
	 * @brief Componente Z del viento
	 */
	double vz;
	/**
	 * @brief Componente X del viento
	 */
	double vx_orig;
	/**
	 * @brief Componente Y del viento
	 */
	double vy_orig;
	/**
	 * @brief Componente Z del viento
	 */
	double vz_orig;
	/**
	 * @brief Boleano que indica si la estación está a 10m o no
	 */
	bool noMismaAltura;

};

/**
 * @brief Estructura para representar el conjunto de las estaciones
 */
struct T_NEstacion{
	/**
	 * @brief Número de estaciones
	 */
	size_t NEST;
	/**
	 * @brief Vector de estaciones
	 */
	struct T_Estacion **N;
	/**
	 * @brief Número de estaciones que intervienen en el cálculo de la función objetivo
	 */
	int Nfo;
	/**
	 * @brief Boleano que indica si todas las estaciones estan a la misma altura o no
	 */
	bool noMismaAltura;
};

/**
 * @brief Vector 3D con sus componentes x,y,z
 */
struct T_Vector3D{
	/**
	 * @brief Componente X del vector
	 */
	double x;
	/**
	 * @brief Componente Y del vector
	 */
	double y;
	/**
	 * @brief Componente Z del vector
	 */
	double z;
};

/**
 * @brief Estructura para guardar datos relativos a las condiciones atmosfericas,
 *        tiempo de simulacion y parametros del modelo
 */
struct T_DatosAtmos{
	/**
	 * @brief Latitud en grados
	 */
	double dlat;
	/**
	 * @brief Velocidad de rotación de la tierra (rad/s)
	 */
	double omega;
	/**
	 * @brief Constante de von Karman
	 */
	double dk;
	/**
	 * @brief Tipo de estabilidad de Pasquill
	 */
	int pasquill;
	/**
	 * @brief Viento geostrófico
	 */
	struct T_Vector3D vg;
	/**
	 * @brief Parámetro para el cálculo de la altura de la capa planetaria
	 */
	double gamma;
	/**
	 * @brief Parámetro para el cálculo de la h en condiciones estables
	 */
	double gamma2;
	/**
	 * @brief Parámetro de Coriolis f=2*omega*sin(dlat)
	 */
	double f;
	/**
	 * @brief Parámetro para dar peso al viento horiz. y vertical
	 */
	double epsilon;
	/**
	 * @brief Parámetro alpha del modelo
	 */
	double alpha;
	/**
	 * @brief Tiempo de simulación
	 */
	int t;
	/**
	 * @brief Altura de las estaciones sobre el terreno
	 */
	double altest;
	/**
	 * @brief Método de resolucion del sistema de ecuaciones
	 */
	int metodo;
	/**
	 * @brief Tipo de precondicionador para el sistema de ecuaciones
	 */
	int precond;
	/**
	 * @brief parametro del precondicionador SSOR
	 */
	double w_prec;
	/**
	 * @brief Número de estaciones
	 */
	size_t nest;
	/**
	 * @brief Nombre del fichero de viento
	 */
	char * fich_viento;
	/**
	 * @brief Nombre del fichero de topografía
	 */
	char *fich_topo;
	/**
	 * @brief fich_malla
	 */
	char *fich_malla;
	/**
	 * @brief Valor que define el viento
	 *        0 : el viento esta en componentes polares (solo se lee el
	 *            modulo y el angulo)
	 *        1 : el viento esta en componentes cartesianas (se lee
	 *            Vx, Vy, Vz)
	 */
	int tipo_fich_viento;
	/**
	 * @brief Variables del harmonie
	 */
	struct VariableHARMONIE *H;
	/**
	 * @brief rugosidad Valor constante de rugosidad
	 */
	double rugosidad;
	/**
	 * @brief tieneRugosidad Booleano para saber si la rugosidad es constante
	 */
	bool tieneRugosidad;
};



/**
 * @brief Función que sirve para crear la estructura malla a partir del fichero de malla
 * @param[in] f_malla Cadena de texto con el nombre del fichero de malla
 * @param[out] malla Estructura con la malla resultante
 */
void
creaMalla(const char *f_malla, struct T_Malla *malla);

/**
 * @brief Función que lee el viento de entrada (tanto de estaciones, modelo de predicción, o interpolado)
 * @param[in] file_format Tipo de viento de entrada (stations, forecast, interpolation)
 * @param[inout] Datmos Variable con los datos atmosféricos. En caso de forecast se rellena la componente H
 * @param[out] todas_est Variable con todas las estaciones. Se rellena en caso de stations
 * @param[out] topo Variable con los valores de rugosidad de la topografia
 * @param[inout] malla Variable con la malla. En el caso interpolation se modifica la componente vini
 */
void leerViento(const char *file_format, struct T_DatosAtmos *Datmos, struct T_NEstacion *todas_est,
		struct T_Topografia *topo, struct T_Malla *malla);

/**
 * @brief Función que a partir de los datos atmosféricos y las estaciones interpola el viento en toda la malla
 * @param[in] Datmos Estructura con los datos atmosféricos
 * @param[inout] malla Estructura con la malla, el viento interpolado se guarda en la componente vini
 * @param[in] topo Estructura con la topografia
 * @param[out] todas_est Estructura con todas las estaciones
 * @param[in] file_format Tipo de fichero de estaciones
 */
void
interpolacionViento(const struct T_DatosAtmos *Datmos, struct T_Malla *malla,
		const struct T_Topografia *topo, struct T_NEstacion *todas_est, const char *file_format);

/**
 * @brief Función que calcula el campo de viento resultante de aplicar masa consistente
 * @param[inout] malla Estructura con la malla, el viento resultante se guarda en la componente wind
 * @param[in] Datmos Estructura con los datos atmosféricos
 */
void
calculoViento(struct T_Malla *malla, const struct T_DatosAtmos *Datmos);

/**
 * @brief Función que escribe los ficheros de resultados
 * @param[in] malla Estructura con la malla
 */
void
escribirResultados(const struct T_Malla *malla);

/**
 * @brief Función que libera la memoria de todas las estructuras usadas en el programa
 * @param[inout] fich_param Cadena del fichero de parametros
 * @param[inout] file_format Cadena del tipo de fichero de estaciones
 * @param[inout] malla Estructura de malla
 * @param[inout] Datmos Estructura de datos atmosféricos
 * @param[inout] topo Estructura de topografia
 * @param[inout] todas_est Estructura de estaciones
 */
void
libera_memoria(char *fich_param, char *file_format, struct T_Malla *malla,
		struct T_DatosAtmos *Datmos, struct T_Topografia *topo,
		struct T_NEstacion *todas_est);

#endif //LECTURA_H
